# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>
    Yevgeniy Yakusevich <eugenethree@gmail.com>
    James Kring <jdk0026@tigermail.auburn.edu>

Entry point to XICSRT.
Contains the main functions that are called to perform raytracing.
"""

import numpy as np

import os

from copy import deepcopy

from xicsrt.util import mirlogging
from xicsrt.util import profiler

from xicsrt import xicsrt_config
from xicsrt import xicsrt_io
from xicsrt.objects._Dispatcher import Dispatcher
from xicsrt.objects._RayArray import RayArray

m_log = mirlogging.getLogger(__name__)

def raytrace(config):
    """
    Perform a series of ray tracing runs.

    Each run will rebuild all objects, reset the random seed and then
    perform the requested number of iterations.

    If the option 'save_images' is set, then images will be saved
    at the completion of each run. The saving of these run images
    is one reason to use this routine rather than just increasing
    the number of iterations: periodic outputs during long computations.

    Also see :func:`~xicsrt.xicsrt_multiprocessing.raytrace` for a
    multiprocessing version of this routine.
    """
    profiler.start('raytrace')
    
    # Update the default config with the user config.
    config = xicsrt_config.get_config(config)
    check_config(config)

    # Make local copies of some options.
    num_runs = config['general']['number_of_runs']
    random_seed = config['general']['random_seed']
    
    output_list = []
    
    for ii in range(num_runs):
        m_log.info('Starting run: {} of {}'.format(ii + 1, num_runs))
        config_run = deepcopy(config)
        config_run['general']['output_run_suffix'] = '{:04d}'.format(ii)

        # Make sure each run uses a unique random seed.
        if random_seed is not None:
            random_seed += ii
        config_run['general']['random_seed'] = random_seed
        
        iteration = raytrace_single(config_run, _internal=True)
        output_list.append(iteration)
        
    output = combine_raytrace(output_list)

    # Reset the configuration options that were unique to the individual runs.
    output['config']['general']['output_run_suffix'] = config['general']['output_run_suffix']
    output['config']['general']['random_seed'] = config['general']['random_seed']

    if config['general']['save_config']:
        xicsrt_io.save_config(output['config'])
    if config['general']['save_images']:
        xicsrt_io.save_images(output)
    if config['general']['save_results']:
        xicsrt_io.save_results(output)
    if config['general']['print_results']:
        print_raytrace(output)

    profiler.stop('raytrace')
    return output


def raytrace_single(config, _internal=False):
    """
    Perform a single raytrace run consisting of multiple iterations.

    If history is enabled, sort the rays into those that are detected and
    those that are lost (found and lost). The found ray history will be
    returned in full. The lost ray history will be truncated to allow
    analysis of lost ray pattern while still limiting memory usage.

    private keywords
    ================
    _internal : bool (False)
      Used when calling this function from `raytrace` as part of the execution
      of multiple runs. Controls how `history_max_lost` is handled along with
      how `save_config` and `save_results` are interpreted.
    """
    profiler.start('raytrace_single')

    # Update the default config with the user config.
    config = xicsrt_config.config_to_numpy(config)
    config = xicsrt_config.get_config(config)
    check_config(config)

    m_log.info('Seeding np.random with {}'.format(config['general']['random_seed']))
    np.random.seed(config['general']['random_seed'])

    num_iter = config['general']['number_of_iter']
    max_lost_iter = int(config['general']['history_max_lost']/num_iter)

    if _internal:
        max_lost_iter = max_lost_iter//config['general']['number_of_runs']

    # Save at least one lost ray even if this exceeds history_max_lost.
    max_lost_iter = max(int(max_lost_iter), 1)

    # Setup the dispatchers.
    if 'filters' in config:
        m_log.debug("Creating filters")
        filters = Dispatcher(config, 'filters')
        filters.instantiate()
        filters.setup()
        filters.initialize()
        config['filters'] = filters.get_config()
    else:
        filters = None

    m_log.debug("Creating sources")
    sources = Dispatcher(config, 'sources')
    sources.instantiate()
    sources.apply_filters(filters)
    sources.setup()
    sources.check_param()
    sources.initialize()
    config['sources'] = sources.get_config()

    m_log.debug("Creating optics")
    optics = Dispatcher(config, 'optics')
    optics.instantiate()
    optics.apply_filters(filters)
    optics.setup()
    optics.check_param()
    optics.initialize()
    config['optics'] = optics.get_config()

    # Do the actual raytracing
    output_list = []
    for ii in range(num_iter):
        m_log.info('Starting iteration: {} of {}'.format(ii + 1, num_iter))

        single = _raytrace_iter(config, sources, optics)
        sorted = _sort_raytrace(single, max_lost=max_lost_iter)
        output_list.append(sorted)

    output = combine_raytrace(output_list)

    if _internal is False:
        if config['general']['print_results']:
            print_raytrace(output)
        if config['general']['save_config']:
            xicsrt_io.save_config(output['config'])
        if config['general']['save_results']:
            xicsrt_io.save_results(output)

    if config['general']['save_images']:
        xicsrt_io.save_images(output)

    profiler.stop('raytrace_single')
    # profiler.report()
    return output


def _raytrace_iter(config, sources, optics):
    """ 
    Perform a single iteration of raytracing with the given sources and optics.
    The returned rays are unsorted.
    """
    profiler.start('_raytrace_iter')

    # Setup local names for a few config entries.
    # This is only to make the code below more readable.
    keep_meta    = config['general']['keep_meta']
    keep_images  = config['general']['keep_images']
    keep_history = config['general']['keep_history']

    m_log.debug('Generating rays')
    rays = sources.generate_rays(keep_history=keep_history)
    m_log.debug('Raytracing optics')
    rays = optics.trace(rays, keep_history=keep_history, keep_images=keep_images)

    # Combine sources and optics outputs.
    meta    = dict()
    image   = dict()
    history = dict()
    
    if keep_meta:
        for key in sources.meta:
            meta[key] = sources.meta[key]
        for key in optics.meta:
            meta[key] = optics.meta[key]
    
    if keep_images:
        for key in sources.image:
            image[key] = sources.image[key]
        for key in optics.image:
            image[key] = optics.image[key]    
    
    if keep_history:
        for key in sources.history:
            history[key] = sources.history[key]
        for key in optics.history:
            history[key] = optics.history[key]

    output = dict()
    output['config'] = config
    output['meta'] = meta
    output['image'] = image
    output['history'] = history
    
    profiler.stop('_raytrace_iter')
    return output


def _sort_raytrace(input, max_lost=None):
    """
    Sort the rays into 'lost' and 'found' rays, then truncate
    the number of lost rays.
    """
    if max_lost is None:
        max_lost = 1000

    profiler.start('_sort_raytrace')

    output = dict()
    output['config'] = input['config']
    output['total'] = dict()
    output['total']['meta'] = dict()
    output['total']['image'] = dict()
    output['found'] = dict()
    output['found']['meta'] = dict()
    output['found']['history'] = dict()
    output['lost'] = dict()
    output['lost']['meta'] = dict()
    output['lost']['history'] = dict()

    output['total']['meta'] = input['meta']
    output['total']['image'] = input['image']

    if len(input['history']) > 0:
        key_opt_list = list(input['history'].keys())
        key_opt_last = key_opt_list[-1]

        w_found = np.flatnonzero(input['history'][key_opt_last]['mask'])
        w_lost = np.flatnonzero(np.invert(input['history'][key_opt_last]['mask']))

        # Save only a portion of the lost rays so that our lost history does
        # not become too large.
        max_lost = min(max_lost, len(w_lost))
        index_lost = np.arange(len(w_lost))
        np.random.shuffle(index_lost)
        w_lost = w_lost[index_lost[:max_lost]]

        for key_opt in key_opt_list:
            output['found']['history'][key_opt] = dict()
            output['lost']['history'][key_opt] = dict()

            for key_ray in input['history'][key_opt]:
                output['found']['history'][key_opt][key_ray] = input['history'][key_opt][key_ray][w_found]
                output['lost']['history'][key_opt][key_ray] = input['history'][key_opt][key_ray][w_lost]

    profiler.stop('_sort_raytrace')

    return output


def combine_raytrace(input_list,
                     keep_images=True,
                     components=None):
    """
    Produce a combined results dictionary from a list of raytrace results.

    Keywords
    --------

    keep_images: bool (True)
        Control whether to combine images.
        Useful when combining outputs which use different optics sizes.

    components: list (None)
        A list of specific components to combine.
        Useful when not all components are needed in the final output.

    Example
    -------
    results_1 = xicsrt.raytrace(config_1)
    results_2 = xicsrt.raytrace(config_2)
    results = xicsrt_raytrace.combine_raytrace([results_1, results_2])
    """
    profiler.start('combine_raytrace')

    output = dict()
    output['config'] = input_list[0]['config']
    output['total'] = dict()
    output['total']['meta'] = dict()
    output['total']['image'] = dict()
    output['found'] = dict()
    output['found']['meta'] = dict()
    output['found']['history'] = dict()
    output['lost'] = dict()
    output['lost']['meta'] = dict()
    output['lost']['history'] = dict()

    num_iter = len(input_list)

    if components is None:
        key_opt_list = list(input_list[0]['total']['meta'].keys())
    else:
        key_opt_list = components

    key_opt_last = key_opt_list[-1]

    # Combine the meta data.
    for key_opt in key_opt_list:
        output['total']['meta'][key_opt] = dict()
        key_meta_list = list(input_list[0]['total']['meta'][key_opt].keys())
        for key_meta in key_meta_list:
            output['total']['meta'][key_opt][key_meta] = 0
            for ii_iter in range(num_iter):
                output['total']['meta'][key_opt][key_meta] += input_list[ii_iter]['total']['meta'][key_opt][key_meta]

    # Combine the images.
    if keep_images:
        for key_opt in key_opt_list:
            if key_opt in input_list[0]['total']['image']:
                if input_list[0]['total']['image'][key_opt] is not None:
                    # Check shape compatibility
                    shape = input_list[0]['total']['image'][key_opt].shape
                    shape_match = True
                    for ii_iter in range(num_iter):
                        if not input_list[ii_iter]['total']['image'][key_opt].shape == shape:
                            shape_match = False
                    if shape_match:
                        # Shapes match, combine images
                        output['total']['image'][key_opt] = np.zeros(input_list[0]['total']['image'][key_opt].shape)
                        for ii_iter in range(num_iter):
                            output['total']['image'][key_opt] += input_list[ii_iter]['total']['image'][key_opt]
                    else:
                        m_log.warning('Image dimensions do not match. Cannot combine images.')
                        output['total']['image'][key_opt] = None
                else:
                    output['total']['image'][key_opt] = None

    # Combine all the histories.
    if len(input_list[0]['found']['history']) > 0:
        final_num_found = 0
        final_num_lost = 0
        for ii_run in range(num_iter):
            final_num_found += len(input_list[ii_run]['found']['history'][key_opt_last]['mask'])
            final_num_lost += len(input_list[ii_run]['lost']['history'][key_opt_last]['mask'])

        rays_found_temp = RayArray()
        rays_found_temp.zeros(final_num_found)

        rays_lost_temp = RayArray()
        rays_lost_temp.zeros(final_num_lost)

        for key_opt in key_opt_list:
            output['found']['history'][key_opt] = rays_found_temp.copy()
            output['lost']['history'][key_opt] = rays_lost_temp.copy()

        index_found = 0
        index_lost = 0
        for ii_run in range(num_iter):
            num_found = len(input_list[ii_run]['found']['history'][key_opt_last]['mask'])
            num_lost = len(input_list[ii_run]['lost']['history'][key_opt_last]['mask'])

            for key_opt in key_opt_list:
                for key_ray in output['found']['history'][key_opt]:
                    output['found']['history'][key_opt][key_ray][index_found:index_found + num_found] = (
                        input_list[ii_run]['found']['history'][key_opt][key_ray][:])
                    output['lost']['history'][key_opt][key_ray][index_lost:index_lost + num_lost] = (
                        input_list[ii_run]['lost']['history'][key_opt][key_ray][:])

            index_found += num_found
            index_lost += num_lost

    profiler.stop('combine_raytrace')
    return output


def check_config(config):
    """
    Check the general section of the configuration dictionary.
    """

    # Check if anything needs to be saved.
    do_save = False
    for key in config['general']:
        if 'save' in key:
            if config['general'][key]:
                do_save = True

    if do_save:
        if not xicsrt_io.path_exists(config['general']['output_path']):
            if not config['general']['make_directories']:
                raise Exception('Output directory does not exist. Create directory or set make_directories to True.')


def print_raytrace(results):
    """
    Print out some information and statistics from the raytracing results.
    """

    key_opt_list = list(results['total']['meta'].keys())
    num_source = results['total']['meta'][key_opt_list[0]]['num_out']
    num_detector = results['total']['meta'][key_opt_list[-1]]['num_out']
    
    print('')
    print('Rays Generated: {:6.3e}'.format(num_source))
    print('Rays Detected:  {:6.3e}'.format(num_detector))
    print('Efficiency:     {:6.3e} ± {:3.1e} ({:7.5f}%)'.format(
        num_detector / num_source,
        np.sqrt(num_detector) / num_source,
        num_detector / num_source * 100))
    print('')




# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>
    Matthew Slominski <mattisaacslominski@yahoo.com>
"""
import numpy as np
from scipy.spatial import Delaunay
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.tools import xicsrt_math as xm

from xicsrt.optics._ShapeMesh import ShapeMesh

@dochelper
class ShapeMeshSphere(ShapeMesh):
    """
    A spherical crystal implemented using a mesh-grid.

    This class meant to be used for three reasons:
    - A spherical optic shape usable with large radii of curvature
    - As an example and template for how to implement a mesh-grid optic.
    - As a verification of the mesh-grid implementation.

    The analytical :class:`ShapeSphere` object should be used for all normal
    raytracing purposes.

    **Programming Notes**

    This optic is built in local coordinates with the mesh surface normal
    generally in the local z = [0, 0, 1] direction and with
    config['trace_local'] = True. This is recommended because the mesh
    implementation performs triangulation and interpolation in the local
    x-y plane.

    """

    def default_config(self):
        """
        radius : float (1.0)
          The radius of the sphere.

        mesh_size : (float, float) ((11,11))
          The number of mesh points in the x and y directions.

        mesh_coarse_size : (float, float) ((5,5))
          The number of mesh points in the x and y directions.
        """
        config = super().default_config()
        config['radius'] = 1.0
        config['mesh_size'] = (11, 11)
        config['mesh_coarse_size'] = (5, 5)

        # The meshgrid is defined in local coordinates.
        config['trace_local'] = True

        return config

    def setup(self):
        super().setup()

        # Generate the fine mesh.
        mesh_points, mesh_normals, mesh_faces = self.generate_mesh(self.param['mesh_size'])
        self.param['mesh_points'] = mesh_points
        self.param['mesh_normals'] = mesh_normals
        self.param['mesh_faces'] = mesh_faces

        # Generate the coarse mesh.
        mesh_points, mesh_normals, mesh_faces = self.generate_mesh(self.param['mesh_coarse_size'])
        self.param['mesh_coarse_points'] = mesh_points
        self.param['mesh_coarse_normals'] = mesh_normals
        self.param['mesh_coarse_faces'] = mesh_faces

    def generate_mesh(self, meshsize):
        """
        Create a spherical meshgrid in local coordinates.
        """
        xsize = self.param['xsize']
        ysize = self.param['ysize']

        x = np.linspace(-xsize/2, xsize/2, meshsize[0])
        y = np.linspace(-ysize/2, ysize/2, meshsize[1])

        # Center of the sphere in local coordinates
        center = np.array([0.0, 0.0, self.param['radius']])

        # Create x,y meshgrid arrays, then calculate z coords
        xx, yy = np.meshgrid(x, y)
        zz = self.param['radius'] - np.sqrt(self.param['radius']**2 - xx**2 - yy**2)

        points = np.stack((xx.flatten(), yy.flatten(), zz.flatten())).T
        norm = xm.normalize(center - points)

        delaunay = Delaunay(points[:, 0:2])
        faces = delaunay.simplices

        return points, norm, faces



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>
"""
import numpy as np

from xicsrt.util import profiler
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._ShapeObject import ShapeObject

from scipy.spatial import Delaunay
from scipy.spatial import cKDTree
from scipy.interpolate import CloughTocher2DInterpolator as Interpolator

@dochelper
class ShapeMesh(ShapeObject):
    """
    A shape that uses a mesh grid instead of an analytical shape.

    This implementation assumes that the mesh surface normal is generally in the
    local z = [0, 0, 1] direction; both triangulation and interpolation are
    done in the x-y plane and are therefore best behaved when this assumption
    holds. It is recommended to build optics in local coordinates with this
    assumption and set config['trace_local'] = True.

    Supplying a well behaved Delaunay Triangulation via config['mesh_delaunay']
    can help to make the mesh work while raytracing in global coordinates,
    but problems will still be encountered if the surface normal is more aligned
    with the x or y directions.

    **Programming Notes**

    Raytracing of mesh optics is fundamentally slow, because of the need to
    find which mesh face is intersected by each ray.  For the simplest
    implementations this requires testing each ray against each mesh face
    leading to the speed scaling as the number of mesh faces
    (or equivalently num_faces^2).

    Some optimization of the basic calculation been completed. The
    mesh_intersect_1 method implements the Möller–Trumbore algorithm and is
    the fastest pure python algorithm found so far.

    To further improve performance this class (optionally) also makes use of
    pre-selection with a coarse mesh. First the intersection with the coarse
    mesh is found for each ray. Then only the 8 nearby faces on the full mesh
    are checked for the final intersection location. This method improves
    the performance to (num_faces_coarse + 8).

    The current algorithm for pre-selection (mesh refinement) is not perfect
    in that the nearby faces are not always appropriately chosen leading to a
    small number of rays being 'lost'. These errors can be minimized by
    increasing the resolution of the coarse mesh and ensuring that the grid
    spacing is approximately equal in the x and y directions.

    Further performance improvement could be gained by using numba or jax.
    This would allow the Möller–Trumbore algorithm to be implemented as a
    loop (instead of being vectorized) where the calculation can be terminated
    early when no hit is found.

    .. Todo::
      XicsrtOpticMesh: Improve the pre-selection (mesh refinement algorithm) to
      eliminate ray losses. The current method is as follows:

      1. Calculate intersection with coarse grid.
      2. Find the point on the fine grid closest to the intersection.
      3. Test all faces on the fine grid that contain this point.

      The problem is that the closest point may not always be part of the face
      that actually has the intersection. This can happen if the fine and coarse
      grid have very different densities, but also even in the perfect case if
      the ray hits near the edge of a face and the grid density is not even in
      the x and y directions.

      What is needed is a better selection of nearby faces. There is also a
      potential to improve computational speed slightly by testing fewer faces
      on the fine grid.
    """

    def default_config(self):
        """
        mesh_points
        mesh_normals
        mesh_faces

        mesh_coarse_points
        mesh_coarse_normals
        mesh_coarse_faces

        mesh_interpolate
        mesh_refine

        """
        config = super().default_config()

        config['mesh_points'] = None
        config['mesh_normals'] = None
        config['mesh_faces'] = None

        config['mesh_coarse_points'] = None
        config['mesh_coarse_normals'] = None
        config['mesh_coarse_faces'] = None

        config['mesh_interpolate'] = None
        config['mesh_refine'] = None

        return config

    def check_param(self):
        super().check_param()

        if self.param['mesh_interpolate'] is None:
            self.param['mesh_interpolate'] = (self.param['mesh_normals'] is not None)
        elif self.param['mesh_interpolate']:
            if self.param['mesh_normals'] is None:
                raise Exception('Surface normal vectors must be defined in order to use mesh interpolation.')

        if self.param['mesh_refine'] is None:
            if self.param['mesh_coarse_points'] is not None:
                self.param['mesh_refine'] = True

        spread_x = np.max(self.param['mesh_points'][:, 0]) - np.min(self.param['mesh_points'][:, 0])
        spread_y = np.max(self.param['mesh_points'][:, 1]) - np.min(self.param['mesh_points'][:, 1])
        spread_z = np.max(self.param['mesh_points'][:, 2]) - np.min(self.param['mesh_points'][:, 2])
        if (spread_z > spread_x) or (spread_z > spread_y):
            self.log.warning('Mesh is not oriented with the surface normals near the local z direction.\n'
                             'This may lead to unexpected and incorrect results.')

    def initialize(self):
        super().initialize()
        self.mesh_initialize()

    def intersect(self, rays):
        """
        Calculate ray intersections with the mesh.
        """
        profiler.start('mesh_intersect')

        if not self.param['mesh_refine']:
            xloc, mask, hits = self.mesh_intersect_1(rays, self.param['mesh'])
        else:
            xloc_c, mask_c, hits_c = self.mesh_intersect_1(
                rays,
                self.param['mesh_coarse'],
                )
            num_rays_coarse = np.sum(mask_c)

            faces_idx, faces_mask = self.find_near_faces(xloc_c, self.param['mesh'], mask_c)
            xloc, mask, hits = self.mesh_intersect_2(
                rays,
                self.param['mesh'],
                mask_c,
                faces_idx,
                faces_mask,
                )
            num_rays_fine = np.sum(mask)

            num_rays_lost = num_rays_coarse - num_rays_fine
            if not num_rays_lost == 0:
                self.log.warning(f'Rays lost in mesh refinement: {num_rays_lost:0.0f} of {num_rays_coarse:6.2e}')

        if self.param['mesh_interpolate']:
            xloc, norm = self.mesh_interpolate(xloc, self.param['mesh'], mask)
        else:
            norm = self.mesh_normals(hits, self.param['mesh'], mask)

        profiler.stop('mesh_intersect')
        return xloc, norm, mask

    def mesh_interpolate(self, X, mesh, mask):
        profiler.start('mesh_interpolate')

        # Interpolate the z-coordinate of the intersection.
        # Here I assume that the surface normal is generally aligned in the
        # z direction. In that case the x and y intersections will be quite
        # accurate, and a correction in the z intersection will help correct
        # for the flatness of each triangle in the triangulation.
        X[:, 2] = mesh['interp']['z'](X[:, 0], X[:, 1])

        normals = np.empty(X.shape)
        normals[:, 0] = mesh['interp']['normal_x'](X[:, 0], X[:, 1])
        normals[:, 1] = mesh['interp']['normal_y'](X[:, 0], X[:, 1])
        normals[:, 2] = mesh['interp']['normal_z'](X[:, 0], X[:, 1])

        profiler.start('normalize')
        normals = np.einsum(
            'i,ij->ij'
            ,1.0/np.linalg.norm(normals, axis=1)
            ,normals
            ,optimize=True)
        profiler.stop('normalize')

        profiler.stop('mesh_interpolate')
        return X, normals

    def _mesh_precalc(self, points, normals, faces):
        profiler.start('_mesh_precalc')

        output = {}
        output['points'] = points
        output['normals'] = normals
        output['faces'] = faces

        # Perform 2D Delaunay triangulation using the x and y locations.
        # This will work fine in most cases, but may cause problems if the
        # optic is oriented so that the normal is in the x or y direction.
        #
        # It is recommended that mesh optics be built using a local
        # coordinate system that makes the x and y coordinates sensible for
        # 2d interpolation.
        delaunay = Delaunay(points[:, 0:2])

        # If pre-triangulated faces were provided, use those for the ray
        # intersection Calculations
        if faces is None:
            faces = delaunay.simplices
            output['faces'] = faces

        if self.param['mesh_interpolate']:
            # Create a set of interpolators.
            profiler.start('Create Interpolators')
            interp = {}
            output['interp'] = interp
            interp['z'] = Interpolator(delaunay, points[:, 2].flatten())
            interp['normal_x'] = Interpolator(delaunay, normals[:, 0].flatten())
            interp['normal_y'] = Interpolator(delaunay, normals[:, 1].flatten())
            interp['normal_z'] = Interpolator(delaunay, normals[:, 2].flatten())
            profiler.stop('Create Interpolators')

        # Copying these makes the code easier to read,
        # but may increase memory usage for dense meshes.
        p0 = points[faces[..., 0], :]
        p1 = points[faces[..., 1], :]
        p2 = points[faces[..., 2], :]

        # Calculate the normals at each face.
        faces_center = np.mean(np.array([p0, p1, p2]), 0)
        faces_normal = np.cross((p0 - p1), (p2 - p1))
        faces_normal /= np.linalg.norm(faces_normal, axis=1)[:, None]
        output['faces_center'] = faces_center
        output['faces_normal'] = faces_normal

        # Generate a tree for the points.
        points_tree = cKDTree(points)
        output['points_tree'] = points_tree

        # Build up a lookup table for the faces around each point.
        # This is currently slow for large arrays.
        points_idx = np.arange(len(points))
        p_faces_idx, p_faces_mask = \
            self.find_point_faces(points_idx, faces)
        output['p_faces_idx'] = p_faces_idx
        output['p_faces_mask'] = p_faces_mask

        # centers_tree = cKDTree(points)
        # output['centers_tree'] = centers_tree

        profiler.stop('_mesh_precalc')
        return output

    def mesh_initialize(self):
        """
        Pre-calculate a number of mesh properties that are
        needed in the other mesh methods.
        """
        profiler.start('mesh_initialize')

        dummy = self._mesh_precalc(
            self.param['mesh_points']
            , self.param['mesh_normals']
            , self.param['mesh_faces'])
        self.param['mesh'] = {}
        for key in dummy:
            self.param['mesh'][key] = dummy[key]

        if self.param['mesh_coarse_points'] is not None:
            dummy = self._mesh_precalc(
                self.param['mesh_coarse_points']
                , self.param['mesh_coarse_normals']
                , self.param['mesh_coarse_faces'])
            self.param['mesh_coarse'] = {}
            for key in dummy:
                self.param['mesh_coarse'][key] = dummy[key]

        profiler.stop('mesh_initialize')

    def mesh_intersect_1(self, rays, mesh):
        """
        Find the intersection of rays with the mesh using the Möller–Trumbore
        algorithm.
        """
        profiler.start('mesh_intersect_1')
        O = rays['origin']
        D = rays['direction']

        m = rays['mask'].copy()
        X = np.full(D.shape, np.nan, dtype=np.float64)

        # Copying these makes the code easier to read,
        # but may increase memory usage for dense meshes.
        p0 = mesh['points'][mesh['faces'][..., 0], :]
        p1 = mesh['points'][mesh['faces'][..., 1], :]
        p2 = mesh['points'][mesh['faces'][..., 2], :]

        epsilon = 1e-15

        num_rays = len(m)
        hits = np.empty(num_rays, dtype=np.int64)
        m_temp = np.empty(num_rays, dtype=bool)
        m_temp_2 = np.zeros(num_rays, dtype=bool)

        for ii in range(mesh['faces'].shape[0]):
            m_temp[:] = m
            edge1 = p1[ii, :] - p0[ii, :]
            edge2 = p2[ii, :] - p0[ii, :]
            h = np.cross(D, edge2)
            f = np.einsum('i,ji->j', edge1, h, optimize=True)
            m_temp &= ~((f > -epsilon) & (f < epsilon))
            if not np.any(m_temp):
                continue

            f = 1.0 / f
            s = O - p0[ii, :]
            u = f * np.einsum('ij,ij->i', s, h, optimize=True)
            m_temp &= ~((u < 0.0) | (u > 1.0))
            if not np.any(m_temp):
                continue

            q = np.cross(s, edge1)
            v = f * np.einsum('ij,ij->i', D, q, optimize=True)
            m_temp &= ~((v < 0.0) | (u + v > 1.0))
            if not np.any(m_temp):
                continue

            t = f * np.einsum('i,ji->j', edge2, q, optimize=True)

            # Update overall hit array and hit mask.
            m_temp_2[m_temp] = m_temp[m_temp]
            hits[m_temp] = ii
            X[m_temp] = O[m_temp] + t[m_temp, None] * D[m_temp, :]

        # Update the mask not to include any rays that didn't hit the mesh.
        m &= m_temp_2
        profiler.stop('mesh_intersect_1')

        return X, m, hits

    def mesh_intersect_2(
            self,
            rays,
            mesh,
            mask,
            faces_idx,
            faces_mask,
            ):
        """
        Check for ray intersection with a list of mesh faces.

        Programming Notes
        -----------------
        Because of the mesh indexing, the arrays here have different
        dimensions than in mesh_intersect_1, and need a different
        vectorization.

        At the moment I am using a less efficient mesh intersection
        method. This should be updated to use the same method as
        mesh_intersect_1, but with the proper vectorization.
        """

        profiler.start('mesh_intersect_2')

        O = rays['origin']
        D = rays['direction']

        m =  mask.copy()
        X = np.full(D.shape, np.nan, dtype=np.float64)

        num_rays = len(m)
        hits = np.empty(num_rays, dtype=np.int64)
        epsilon = 1e-15

        # Copying these makes the code easier to read,
        # but may increase memory usage for dense meshes.
        faces = mesh['faces'][faces_idx]
        p0 = mesh['points'][faces[..., 0], :]
        p1 = mesh['points'][faces[..., 1], :]
        p2 = mesh['points'][faces[..., 2], :]
        n = mesh['faces_normal'][faces_idx]

        # distance = np.dot((p0 - O), n) / np.dot(D, n)
        t0 = p0[:, :, :] - O[None, :, :]
        t1 = np.einsum('ijk, ijk -> ij', t0, n, optimize=True)
        t2 = np.einsum('jk, ijk -> ij', D, n, optimize=True)
        dist = t1 / t2
        t3 = np.einsum('jk,ij -> ijk', D, dist, optimize=True)
        intersect = t3 + O

        # Pre-calculate some vectors and do the calculation in a
        # single step in an attempt to optimize this calculation.
        a = intersect - p0
        b = intersect - p1
        c = intersect - p2

        diff = (np.linalg.norm(np.cross(b, c), axis=2)
                + np.linalg.norm(np.cross(c, a), axis=2)
                + np.linalg.norm(np.cross(a, b), axis=2)
                - np.linalg.norm(np.cross((p0 - p1), (p0 - p2)), axis=2)
                )

        # .. ToDo:
        #    For now hard code the floating point tolerance.
        #    A better way of handling floating point errors is needed.
        test = (diff < 1e-10) & (dist >= 0) & faces_mask
        m &= np.any(test, axis=0)

        # idx_hits tells us which of the 8 faces had a hit.
        idx_hits = np.argmax(test[:, m], axis=0)
        # Now index the faces_idx to git the actual face number.
        hits[m] = faces_idx[idx_hits, m]
        X[m] = intersect[idx_hits, m, :]

        profiler.stop('mesh_intersect_2')

        return X, m, hits

    def mesh_normals(self, hits, mesh, mask):
        m = mask
        normals = np.zeros((len(m), 3), dtype=np.float64)
        normals[m, :] = mesh['faces_normal'][hits[m], :]
        return normals

    def mesh_get_index(self, hits, faces):
        """
        Match faces to face indexes, with a loop over faces.
        """
        profiler.start('mesh_get_index')
        idx_hits = np.empty(hits.shape[0], dtype=np.int32)
        for ii, ff in enumerate(faces):
            m_temp = np.all(np.equal(ff, hits), axis=1)
            idx_hits[m_temp] = ii
        profiler.stop('mesh_get_index')
        return idx_hits

    def find_point_faces(self, p_idx, faces, mask=None):
        """
        Find all of the the faces that include a given mesh point.
        """
        profiler.start('find_point_faces')
        if mask is None:
            mask = np.ones(p_idx.shape, dtype=np.bool_)
        m = mask
        p_faces_idx = np.zeros((8, len(m)), dtype=np.int32)
        p_faces_mask = np.zeros((8, len(m)), dtype=np.bool_)
        for ii_p in p_idx:
            ii_f = np.nonzero(np.equal(faces, p_idx[ii_p]))[0]
            faces_num = len(ii_f)
            p_faces_idx[:faces_num, ii_p] = ii_f
            p_faces_mask[:faces_num, ii_p] = True
        profiler.stop('find_point_faces')
        return p_faces_idx, p_faces_mask

    def find_near_faces(self, X, mesh, mask):
        m = mask
        profiler.start('find_near_faces')
        idx = mesh['points_tree'].query(X[m])[1]

        faces_idx = np.zeros((8, len(m)), dtype=np.int32)
        faces_mask = np.zeros((8, len(m)), dtype=np.bool_)

        faces_idx[:, m] = mesh['p_faces_idx'][:, idx]
        faces_mask[:, m] = mesh['p_faces_mask'][:, idx]
        profiler.stop('find_near_faces')
        return faces_idx, faces_mask



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir pablant <npablant@pppl.gov>

A set of tools to help with automatic API documentation of XICSRT.

Description
-----------
XICSRT uses sphinx for documentation, and API docs are based on the idea of
code self documentation though `python` doc strings. This module contains
a set of decorators and helper function to aid in self documentation.

The most important part of this module is the `@dochelper` decorator which
should be used for all element classes.

.. Note:
    Philosophy: Python help() should be just as readable as the Sphinx webpage.

Todo:
    - The config docstrings should all be indented follow the help() standard.
    - Would it be helpful to show which inherited class the options came from?
"""

import inspect
import re


def dochelper(cls):
    """
    A functional wrapper for the DocHelper class. Intended to be used
    as a decorator.

    This decorator does the following:

    1. Adds a 'Configuration Options' section to the class docstring that
       contains all options defined in default_config and any class ancestors.

    """
    return DocHelper(cls)()

class DocHelper:
    """
    A class to help generate docstrings for XICSRT.

    This is expected to be used through the `@dochelper` class decorator.
    """

    def __init__(self, cls):
        self.cls = cls
        self.update_class_docstring(cls)

    def __call__(self):
        return self.cls

    def update_class_docstring(self, cls):
        """
        Update the class docstring. This method does the following:
        1. Creates a new section 'Configuration Options' which contains
           combined documentation for all config options defined in any
           ancestor.
        """
        flag_sort_alpha = False

        if cls.__doc__ == None:
            cls.__doc__ = ''

        sections = {}
        mro = list(inspect.getmro(cls))
        # The order here is important. As written, this will always keep
        # the section from the outermost class, while also adding sections
        # from outermost to innermost.
        for ancestor in mro:
            if hasattr(ancestor, 'default_config'):
                doc = ancestor.default_config.__doc__
                if doc is not None:
                    doc = inspect.getdoc(ancestor.default_config)
                    new = self._parse_config_doc(doc)

                    # I can't use update() because I don't want to overwrite values.
                    for key in new:
                        if key not in sections:
                            sections[key] = new[key]

        # Build up the new config docs from the section dictionary
        new_doc = ''
        key_list = list(sections.keys())
        if flag_sort_alpha:
            key_list.sort()
        for key in key_list:
            new_doc += sections[key]['name']
            if sections[key]['sig']:
                new_doc += ' : ' + sections[key]['sig']
            new_doc += '\n'
            new_doc += sections[key]['body']
            new_doc += '\n'

        cls.__doc__ += '\n'
        cls.__doc__ += 'Configuration Options:\n\n'
        cls.__doc__ += new_doc

    def _parse_config_doc(self, doc):
        """
        Parse configuration documentation into sections. This assumes that
        the config docs are well formatted.
        """
        section_dict = {}
        name = None
        lines = doc.splitlines()
        for line in lines:
            m = re.match(r'^([\w]+)\s*:?\s*(.*)', line)
            if m:
                name = m.group(1)
                section_dict[name] = {}
                section_dict[name]['name'] = m.group(1)
                section_dict[name]['sig'] = m.group(2)
                section_dict[name]['body'] = ''
            else:
                if name:
                    if line.strip():
                        section_dict[name]['body'] += line + '\n'

        return section_dict



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`XicsrtOpticPlanarMosaicCrystal` class.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractMosaicCrystal import InteractMosaicCrystal
from xicsrt.optics._ShapeSphere import ShapeSphere

@dochelper
class XicsrtOpticSphericalMosaicCrystal(InteractMosaicCrystal, ShapeSphere):
    """
    A shperical mosaic crystal optic.
    """

    pass



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`XicsrtOpticDetector` class.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractNone import InteractNone
from xicsrt.optics._ShapePlane import ShapePlane

@dochelper
class XicsrtOpticDetector(InteractNone, ShapePlane):
    """
    A detector optic.

    Programming Notes
    -----------------
    For now the detector class simply records intersections with a plane.
    In the future this class may be expanded to include effects such as quantum
    efficiency, readout noise, dark noise, etc.
    """

    pass



# -*- coding: utf-8 -*-
"""
Authors
-------
- Novimir A. Pablant <npblant@pppl.gov>

"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.filters._XicsrtBundleFilter import XicsrtBundleFilter

@dochelper
class XicsrtBundleFilterSightline(XicsrtBundleFilter):
    """
    A bundle filter based on proximity to sightline vectors.
    """
        
    def default_config(self):
        """
        radius
          The radius of the cylindrical sightline.
        """
        config = super().default_config()

        config['radius'] = None

        return config

    def filter(self, bundle_input):
        """
        Filter ray bundles that do not originate inside the cylindrical
        sightline.
        """

        # vector from sightline origin to bundle position.
        l_0 = self.config['origin'] - bundle_input['origin']
        
        # Projection of l_0 onto the sightline
        proj = np.einsum('j,ij->i', self.config['zaxis'], l_0)[np.newaxis]
        l_1  = np.dot(np.transpose(self.config['zaxis'][np.newaxis]), proj)
        l_1  = np.transpose(l_1)
        
        # Component of l_0 perpendicular to the sightline
        l_2 = l_0 - l_1
        
        # Sightline distance is the length of l_2
        distance = np.sqrt(np.einsum('ij,ij->i', l_2, l_2))
        
        # Check to see if the bundle is close enough to the sightline
        mask = (self.config['radius'] >= distance)
        
        bundle_input['mask'] &= mask
        
        return bundle_input



# -*- coding: utf-8 -*-
"""
Authors
-------
  - Novimir A. Pablant <nablant@pppl.gov>
  - Yevgeniy Yakusevich <eugenethree@gmail.com>
"""

import logging
import numpy as np

from xicsrt.util import profiler
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.sources._XicsrtPlasmaGeneric import XicsrtPlasmaGeneric


@dochelper
class XicsrtPlasmaCubic(XicsrtPlasmaGeneric):
    """
    A cubic plasma.
    """
                
    def bundle_generate(self, bundle_input):

        bundle_input = super().bundle_generate(bundle_input)

        #evaluate temperature at each point
        #plasma cube has consistent temperature throughout
        bundle_input['temperature'][:] = self.param['temperature']
        
        #evaluate emissivity at each point
        #plasma cube has a constant emissivity througout.
        bundle_input['emissivity'][:] = self.param['emissivity']
            
        return bundle_input



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    Nathan Bartlett <nbb0011@auburn.edu>
"""
import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractNone import InteractNone
from xicsrt.optics._ShapePlane import ShapePlane

@dochelper

class XicsrtOpticAperture(InteractNone, ShapePlane):
    """
    An optic that can be used to set an aperture.

    Programming Notes
    -----------------
    All of the implementation for the aperture is in :any:`TraceObject`,
    so nothing is needed in this subclass for the time being.
    """

    pass




# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

Define the :class:`InteractCrystal` class.
"""
import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper

from xicsrt.tools import xicsrt_bragg
from xicsrt.tools import xicsrt_math as xm
from xicsrt.optics._InteractMirror import InteractMirror

@dochelper
class InteractCrystal(InteractMirror):
    """
    Model for simple Bragg reflections.
    """

    def default_config(self):
        """
        crystal_spacing: float [angstroms]
          The spacing between crystal planes.

          .. Note::

            This is the nominal 'd' crystal spacing, not the '2d' spacing
            often used in the literature.

        reflectivity: float (1.0)
          A reflectivity factor for this optic. The reflectivity will modify
          the probability that a ray will reflect from this optic.

        check_bragg: bool (True)
          Switch between x-ray Bragg reflections and optical reflections for
          this optic. If True, a rocking curve will be used to determine the
          probability of reflection for rays based on their incident angle.
          If false this optic will act as a perfect mirror.

        rocking_type: str ('gaussian')
          The type of shape to use for the crystal rocking curve.
          Allowed types are 'step', 'gaussian' and 'file'.

        rocking_fwhm: float [rad]
          The width of the rocking curve, in radians.
          This option only used when rocking_type is 'step' or 'gaussian'.

        rocking_file: str or list
          A filename from which to read rocking curve data. A list may be used
          if sigma and pi data are in separate files.

        rocking_filetype: str
          The type of rocking curve file to be loaded.
          The following formats are currently supported: 'xop', 'x0h', 'simple'.

          .. Note::

            Actually at this point only 'xop' is supported. np 2020-10-13

        rocking_mix: float
          A mixing factor to combine the sigma and pi reflectivities.
          This value will be interpreted as sigma/pi and will mix the reflection
          probabilities linearly. ref = sigma*mix + pi*(1-mix)

        """

        config = super().default_config()
        
        # xray optical information and polarization information
        config['crystal_spacing'] = 0.0
        config['reflectivity']    = 1.0

        config['check_bragg']     = True
        config['rocking_type']       = 'gaussian'
        config['rocking_fwhm']       = None
        config['rocking_file']       = None
        config['rocking_filetype']   = None
        config['rocking_mix']        = 0.5 

        return config
    
    def initialize(self):
        super().initialize()
        self.param['rocking_type'] = str.lower(self.param['rocking_type'])

    def interact(self, rays, xloc, norm, mask=None):
        mask = self.angle_check(rays, norm, mask)
        rays = self.reflect_vectors(rays, xloc, norm, mask)
        rays['mask'] = mask
        return rays

    def angle_calc(self, rays , norm, mask=None):
        if mask is None:
            mask = rays['mask']

        D = rays['direction']
        W = rays['wavelength']
        m = mask

        bragg_angle = np.zeros(m.shape, dtype=np.float64)
        dot = np.zeros(m.shape, dtype=np.float64)
        incident_angle = np.zeros(m.shape, dtype=np.float64)

        # returns vectors that satisfy the bragg condition
        # only perform check on rays that have intersected the optic
        bragg_angle[m] = np.arcsin(W[m] / (2 * self.param['crystal_spacing']))
        dot[m] = np.abs(np.einsum('ij,ij->i', D[m], -1 * norm[m], optimize=True))
        incident_angle[m] = (np.pi / 2) - np.arccos(dot[m] / xm.magnitude(D[m]))

        return bragg_angle, incident_angle

    def angle_check(self, rays, norm, mask=None):
        if mask is None:
            mask = rays['mask']

        if self.param['check_bragg'] is False:
            return mask

        m = mask

        bragg_angle, incident_angle = self.angle_calc(rays, norm, mask)

        #check which rays satisfy bragg, update mask to remove those that don't
        m[m] &= self.rocking_curve_filter(incident_angle[m], bragg_angle[m])

        #print(bragg_angle[0:10])
        #print(incident_angle[0:10])
        #print(mask[0:10])

        return mask

    def rocking_curve_filter(self, incident_angle, bragg_angle):

        # Generate or load a probability curve.
        if "step" in self.param['rocking_type']:
            # Step Function
            p = np.where(abs(incident_angle - bragg_angle) <= self.param['rocking_fwhm'] / 2,
                         1.0, 0.0)

        elif "gauss" in self.param['rocking_type']:
            # Convert from FWHM to sigma.
            sigma = self.param['rocking_fwhm'] / (2 * np.sqrt(2 * np.log(2)))

            # Gaussian Distribution
            p = np.exp(-np.power(incident_angle - bragg_angle, 2.) / (2 * sigma**2))

        elif "file" in self.param['rocking_type']:

            data = xicsrt_bragg.read(self.param['rocking_file'], self.param['rocking_filetype'])

            units = data['units']['dtheta_in']
            if units == 'urad':
                scale_theta = 1e-6
            elif units == 'arcset':
                scale_theta = np.pi / (180 * 3600)
            elif units == 'rad':
                scale_theta = 1.0
            else:
                raise Exception(f'Units in rocking curve data not understood: {units}')

            # evaluate rocking curve reflectivity value for each incident ray
            # This calculation is done in units of 'radians'.
            dtheta = (incident_angle - bragg_angle)
            sigma = np.interp(
                dtheta
                , data['value']['dtheta_in'] * scale_theta
                , data['value']['reflect_s']
                , left=0.0
                , right=0.0)
            pi = np.interp(
                dtheta
                , data['value']['dtheta_in'] * scale_theta
                , data['value']['reflect_p']
                , left=0.0
                , right=0.0)

            p = self.param['rocking_mix'] * sigma + (1 - self.param['rocking_mix']) * pi

        else:
            raise Exception('Rocking curve type not understood: {}'.format(self.param['rocking_type']))

        p *= self.param['reflectivity']

        # Rreate a random number for each ray between 0 and 1.
        test = np.random.uniform(0.0, 1.0, len(incident_angle))

        # compare that random number with the reflectivity distribution
        # and filter out the rays that do not survive the random
        # rocking curve filter.
        mask = (p >= test)

        return mask



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`ShapeTorus` class.
"""
import numpy as np

from sys import float_info
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.tools import xicsrt_math as xm
from xicsrt.tools import xicsrt_quartic as xq
from xicsrt.optics._ShapeObject import ShapeObject

@dochelper
class ShapeTorus(ShapeObject):
    """
    A Toroidal shape.
    This class defines intersections with a torus.
    """

    def default_config(self):
        config = super().default_config()
        
        """
        radius_major: float (1.0)
            The radius of curvature of the crystal in the toroidal (xaxis)
            direction. This is not the same as the geometric major radius of the
            axis of a toroid, which in our case would be r_major-r_minor.

        radius_minor: float (0.1)
            The radius of the curvature of the crystal in the poloidal (yaxis)
            direction. This is the same as the geometric minor radius of a
            toroid.
            
        convex : [bool, bool] ([False, False])
            Whether the major and minor curvature will be convex or concave.
            This has the effect of determining which of the four possible 
            intersections with the torus will be chosen. For a ray starting
            at infinity and passing through all four possible intersections
            with a torus, then the intersections correspond to:
              1 -> [True, True]
              2 -> [True, False]
              3 -> [False, False]
              4 -> [False, False]
        """
        
        config['radius_major']  = 1.0
        config['radius_minor']  = 0.2
        config['convex'] = [False, False]

        return config

    def initialize(self):
        super().initialize()
        
        r_minor = self.param['radius_minor']
        r_major = self.param['radius_major']

        if r_minor >= r_major:
            raise Exception(r'Cannot construct geometry with radius_major <= radius_minor.')

        # radius_major is the radius of curvature at the surface; here we find
        # the major and minor radii for the geometrical toroid construction
        # along with the geometrical center (in global coordinates).
        #
        # The indexing of the roots are specific to the algebraic quartic
        # solver in xicsrt_quartic.

        self.param['torus_minor'] = r_minor

        if np.all(self.param['convex'] == [False, False]):
            self.param['root_idx'] = 3
            self.param['torus_major'] = r_major - r_minor
            sign = 1
        elif np.all(self.param['convex'] == [False, True]):
            self.param['root_idx'] = 2
            self.param['torus_major'] = r_major + r_minor
            sign = 1
        elif np.all(self.param['convex'] == [True, False]):
            self.param['root_idx'] = 1
            self.param['torus_major'] = r_major + r_minor
            sign = -1
        elif np.all(self.param['convex'] == [True, True]):
            self.param['root_idx'] = 0
            self.param['torus_major'] = r_major - r_minor
            sign = -1
        else:
            raise Exception(f"Cannot be parse convex config option: {self.param['convex']}")

        self.param['center'] = self.param['origin'] + sign*r_major*self.param['zaxis']


    def intersect(self, rays):
        dist, mask = self.intersect_distance(rays)
        xloc = self.location_from_distance(rays, dist, mask)
        norm = self.intersect_normal(xloc, mask)

        return xloc, norm, mask


    def point_to_toroid(self, point_external, copy=False):
        """
        Convert a point into 'toroid' coordinates. In this coordinate system the
        origin is at the center of the geometrical torus.
        """
        return self.vector_to_local(point_external - self.param['center'], copy=copy)


    def intersect_distance(self, rays):
        """
        Calculate the distance to the intersection of the rays with the
        Toroidal optic.

        Programming Notes
        -----------------

        This calculation is done in 'torus' coordinates in which the center of
        the geometric torus is the origin.  The rays are converted to this
        coordinate system it the beginning of this method. Since we are only
        calculating the distance here, we don't need to convert back.
        """
        
        # Create short variable names for readability.
        O = rays['origin']
        D = rays['direction']
        m = rays['mask']

        r_minor = self.param['torus_minor']
        r_major = self.param['torus_major']

        O = self.point_to_toroid(O, copy=True)
        D = self.vector_to_local(D, copy=True)
        
        # defining resusable variables
        O_mag_sq = np.einsum('ij,ij->i', O, O)
        dot_OD = np.einsum('ij,ij->i', O, D)

        r_sq = r_major ** 2 + r_minor ** 2

        # The form of quartic equation to be solved is,
        # c0 ** t ^ 4 + c1 ** t ^ 3 + c2 ** t ^ 2 + c3 ** t + c4 = 0

        # Define the Axis of the toroid to be the y-axis.
        # 0 - x-axis, 1 - y-axis, 2 - z-axis
        axis_idx = 1
        ones = np.ones(len(m))

        # defining co-efficients
        # Coefficient of t^4
        c0 = ones**4
        # Coefficient of t^3
        c1 = 4 * ones**2 * dot_OD
        # Coefficient of t^2
        c2 = 4 * dot_OD**2 + 2 * O_mag_sq * ones**2 - 2 * r_sq * ones**2 + 4 * r_major**2 * D[:, axis_idx]**2
        # Coefficient of t^1
        c3 = 4 * dot_OD * (O_mag_sq - r_sq) + 8 * r_major**2 * D[:, axis_idx] * O[:, axis_idx]
        # Coefficient of t^0
        c4 = O_mag_sq**2 - 2 * r_sq * O_mag_sq + 4 * r_major**2 * O[:, axis_idx]**2 + (r_major**2 - r_minor**2)**2
    
        roots_0, roots_1, roots_2, roots_3 = xq.multi_quartic(c0, c1, c2, c3, c4)
        
        # neglecting complex & negative solution of the quartic equation    
        roots_0[roots_0.imag != 0] = np.nan
        roots_1[roots_1.imag != 0] = np.nan
        roots_2[roots_2.imag != 0] = np.nan
        roots_3[roots_3.imag != 0] = np.nan

        roots = np.stack((
            roots_0.real,
            roots_1.real,
            roots_2.real,
            roots_3.real), axis=1)

        # If we used xics_quartic.multi_quartic, then the roots will always
        # be in the expected order. If this solver gets replaced, some extra
        # logic may be necessary.
        # roots = np.sort(roots, axis=1)

        distances = roots[:, self.param['root_idx']]
        m[m] &= np.isfinite(distances[m]) & (distances[m] > 0.0)

        return distances, m
    

    def intersect_normal(self, xloc, mask):
        """
        Calculate the surface normal at each ray intersection point.

        Programming Notes
        -----------------

        This calculation is done in global coordinates.
        """
        m = mask
        normals = np.zeros(xloc.shape, dtype=np.float64)

        C = self.param['center']
        yaxis = np.cross(self.param['zaxis'], self.param['xaxis'])
        r_major     = self.param['torus_major']

        pt = xloc[m] - C

        # Projection of direction to geometrical center on xz plane.
        pt = pt - np.einsum('i,j->ij', np.einsum('ij,j->i', pt, yaxis), yaxis)
        pt_norm = xm.normalize(pt)

        # Location of the ray aiming point on toroidal axis.
        Q = C + r_major * pt_norm

        X_norm = (xloc[m] - Q)
        X_norm = xm.normalize(X_norm)

        normals[m] = X_norm
        
        return normals 



# -*- coding: utf-8 -*-
"""
..  Authors:
    Novimir Antoniuk Pablant <npablant@pppl.gov>


A slightly more complicated example with an x-ray Bragg reflection from a
spherical crystal.

This configuration has a point source, a spherical crystal, and a detector.
"""

import numpy as np
import xicsrt
xicsrt.warn_version('0.8')

# 1.
config = dict()

# 2.
config['general'] = {}
config['general']['number_of_iter'] = 5
config['general']['save_images'] = False

# 3.
config['sources'] = {}
config['sources']['source'] = {}
config['sources']['source']['class_name'] = 'XicsrtSourceDirected'
config['sources']['source']['intensity'] = 1e4
config['sources']['source']['wavelength'] = 3.9492
config['sources']['source']['spread'] = np.radians(10.0)
config['sources']['source']['xsize'] = 0.00
config['sources']['source']['ysize'] = 0.00
config['sources']['source']['zsize'] = 0.00

# 4.
config['optics'] = {}
config['optics']['crystal'] = {}
config['optics']['crystal']['class_name'] = 'XicsrtOpticCrystalSpherical'
config['optics']['crystal']['check_size'] = True
config['optics']['crystal']['origin'] = [0.0, 0.0,         0.80374151]
config['optics']['crystal']['zaxis']  = [0.0, 0.59497864, -0.80374151]
config['optics']['crystal']['xsize']  = 0.2
config['optics']['crystal']['ysize']  = 0.2
config['optics']['crystal']['radius'] = 1.0

# Rocking curve FWHM in radians.
# This is taken from x0h for quartz 1,1,-2,0
# Darwin Curve, sigma: 48.070 urad
# Darwin Curve, pi:    14.043 urad
config['optics']['crystal']['crystal_spacing'] = 2.45676
config['optics']['crystal']['rocking_type'] = 'gaussian'
config['optics']['crystal']['rocking_fwhm'] = 48.070e-6

# 5.
config['optics']['detector'] = {}
config['optics']['detector']['class_name'] = 'XicsrtOpticDetector'
config['optics']['detector']['origin'] = [0.0,  0.76871290, 0.56904832]
config['optics']['detector']['zaxis']  = [0.0, -0.95641806, 0.29200084]
config['optics']['detector']['xsize']  = 0.4
config['optics']['detector']['ysize']  = 0.2

# 6.
results = xicsrt.raytrace(config)


# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

Description
-----------
Handle reading and writing of files for XICSRT.
"""

import numpy as np
import logging
from copy import deepcopy
import os
import pathlib

import json

from xicsrt import xicsrt_config
from xicsrt.tools import xicsrt_misc
from xicsrt.util import profiler

log = logging.getLogger(__name__)

def load_config(filename):
    config = _dict_from_file(filename)
    return config


def save_config(config, filename=None, path=None, mkdir=None, overwrite=None):
    if filename is None:
        filename = generate_filename(config, kind='config', path=path)
    elif path is not None:
        filename = os.path.join(path, filename)

    if mkdir is None:
        mkdir = config['general'].get('make_directories', False)
    _file_from_dict(config, filename, mkdir=mkdir, overwrite=overwrite)

    filename = pathlib.Path(filename).expanduser().resolve()
    log.info('Config saved to {}'.format(filename))


def save_results(output, filename=None, path=None, mkdir=None, overwrite=None):
    config = output['config']
    if filename is None:
        filename = generate_filename(config, kind='results', path=path)
    elif path is not None:
        filename = os.path.join(path, filename)

    if mkdir is None:
        mkdir = config['general'].get('make_directories', False)
    _file_from_dict(output, filename, mkdir=mkdir, overwrite=overwrite)

    filename = pathlib.Path(filename).expanduser().resolve()
    log.info('History saved to {}'.format(filename))


def load_results(filename=None, path=None, config=None):
    """
    Load an XICSRT results file.

    Keywords
      filename
        The name of the file to load.
        filename can be a full path; otherwise if path is given it will be
        prepended to the filename.
        If filename is not provided and a config is given, then a filename
        will be auto generated based on the config.
      path
        If provided the path will be prepended to the filename.
      config
        If provided (and filename is not), then the filename will be auto
        generated using the config settings.
    """
    print(path, filename)
    if filename is None:
        filename = generate_filename(config, kind='results', path=path)
    elif path is not None:
        filename = os.path.join(path, filename)

    print(filename)

    output = _dict_from_file(filename)
    return output


def save_images(output, rotate=True, path=None, mkdir=None):
    """
    Save images from the raytracing output.
    """
    from PIL import Image

    config = output['config']
    if mkdir is None:
        mkdir = config['general'].get('make_directories', False)

    for key_opt in output['config']['optics']:
        if key_opt in output['total']['image']:
            if output['total']['image'][key_opt] is not None:
                filename = generate_filename(output['config'], 'image', key_opt, path=path)
                if mkdir:
                    _make_path(filename)

                image_temp = output['total']['image'][key_opt]
                if rotate:
                    image_temp = np.rot90(image_temp)

                generated_image = Image.fromarray(image_temp)
                generated_image.save(filename)

                filename = pathlib.Path(filename).expanduser().resolve()
                log.info('Saved image: {}'.format(filename))


def generate_filename(config, kind=None, name=None, path=None):
    config = xicsrt_config.get_config(config)
    prefix = config['general']['output_prefix']
    suffix = config['general']['output_suffix']
    run_suffix = config['general']['output_run_suffix']
    if path is None:
        path = config['general']['output_path']

    if kind is None:
        ext = ''
    elif kind == 'image':
        ext = config['general']['image_ext']
    elif kind == 'results':
        ext = config['general']['results_ext']
    elif kind == 'config':
        ext = config['general']['config_ext']
    else:
        raise Exception(f'Data kind {kind} unknown.')

    if name is None:
        name = kind

    filename = '_'.join(filter(None, (prefix, name, suffix, run_suffix))) + ext
    filepath = os.path.join(path, filename)

    return filepath


def _dict_from_file(filename):
    filename = pathlib.Path(filename).expanduser()
    ext = filename.suffix

    if ('pickle' in ext) or ('pkl' in ext):
        import pickle
        with open(filename, "rb") as ff:
            data = pickle.load(ff)

    elif ('json' in ext):
        import json
        with open(filename, "r") as ff:
            data = json.load(ff)
        data = xicsrt_misc._convert_to_numpy(data)

    elif ('hdf5' in ext) or ('h5' in ext):
        from xicsrt.util import mirhdf5
        data = mirhdf5.hdf5ToDict(filename)

    else:
        raise NotImplementedError(f'filetype: {ext} not currently supported.')

    return data


def _file_from_dict(data, filename, mkdir=False, overwrite=False):
    if mkdir:
        _make_path(filename)

    filename = pathlib.Path(filename).expanduser()
    if not overwrite and filename.exists():
        raise FileExistsError("File exists. Use overwrite=True to overwrite.")

    ext = filename.suffix

    if ('pickle' in ext) or ('pkl' in ext):
        import pickle
        with open(filename, "wb") as ff:
            pickle.dump(data, ff)

    elif ('json' in ext):
        import json
        data = deepcopy(data)
        data = xicsrt_misc._convert_from_numpy(data)
        with open(filename, "w") as ff:
            json.dump(data, ff, indent=2)

    elif ('hdf5' in ext) or ('h5' in ext):
        from xicsrt.util import mirhdf5
        mirhdf5.dictToHdf5(data, filename)

    else:
        raise NotImplementedError(f'filetype: {ext} not currently supported.')


def _make_output_path(config):
    if config['general'].get('make_directories', False):
        path = pathlib.Path(config['general']['output_path']).expanduser()
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
            log.info(f'Made directory: {path}')


def _make_path(filename):
    path = pathlib.Path(filename).expanduser()
    if path.suffix:
        path = path.parent
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
        log.info(f'Made directory: {path}')


def path_exists(path):
    path = pathlib.Path(path).expanduser()
    return path.exists()



# -*- coding: utf-8 -*-
"""
.. Authors:
   Novimir Pablant <npablant@pppl.gov>
   James Kring <jdk0026@tigermail.auburn.edu>
   Yevgeniy Yakusevich <eugenethree@gmail.com>

Define the :class:`ShapeSpherical` class.
"""
import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.tools import xicsrt_math as xm
from xicsrt.optics._ShapeObject import ShapeObject

@dochelper
class ShapeSphere(ShapeObject):
    """
    A shperical shape.
    This class defines intersections with a sphere.
    """

    def default_config(self):
        """
        radius : float (1.0)
          The radius of the sphere.

        convex : bool (False)
          If True the optic will have a convex curvature, if False the surface
          will have a concave curvature (the default).
        """
        config = super().default_config()
        config['radius'] = 1.0
        config['convex'] = False
        return config

    def initialize(self):
        super().initialize()
        if self.param['convex']:
            sign = -1
        else:
            sign = 1
        self.param['center'] = sign*self.param['radius'] * self.param['zaxis'] + self.param['origin']

    def intersect(self, rays):
        dist, mask = self.intersect_distance(rays)
        xloc = self.location_from_distance(rays, dist, mask)
        norm = self.intersect_normal(xloc, mask)

        return xloc, norm, mask

    def intersect_distance(self, rays):
        """
        Calulate the distance to the intersection of the rays with the
        spherical optic.

        This calculation is copied from:
        https://www.scratchapixel.com/lessons/3d-basic-rendering/minimal-ray-tracer-rendering-simple-shapes/ray-sphere-intersection
        """
        #setup variables
        O = rays['origin']
        D = rays['direction']
        m = rays['mask']
        
        distance = np.full(m.shape, np.nan, dtype=np.float64)
        d        = np.empty(m.shape, dtype=np.float64)
        t_hc     = np.empty(m.shape, dtype=np.float64)
        t_0      = np.empty(m.shape, dtype=np.float64)
        t_1      = np.empty(m.shape, dtype=np.float64)

        # L is the distance from the ray origin to the center of the sphere.
        # t_ca is the projection of this distance along the ray direction.
        L     = self.param['center'] - O
        t_ca  = np.einsum('ij,ij->i', L, D)
        
        # If t_ca is less than zero, then there is no intersection in the
        # the direction of the ray (there might be an intersection behind.)
        # Use mask to only perform calculations on rays that hit the crystal
        # m[m] &= (t_ca[m] >= 0)
        
        # d is the minimum distance between a ray and center of curvature.
        d[m] = np.sqrt(np.einsum('ij,ij->i',L[m] ,L[m]) - t_ca[m]**2)

        # If d is larger than the radius, the ray misses the sphere.
        m[m] &= (d[m] <= self.param['radius'])
        
        # t_hc is the distance from d to the intersection points.
        t_hc[m] = np.sqrt(self.param['radius']**2 - d[m]**2)
        
        t_0[m] = t_ca[m] - t_hc[m]
        t_1[m] = t_ca[m] + t_hc[m]

        # Distance traveled by the ray before hitting the optic surface with
        # the chosen curvature.
        if self.param['convex']:
            distance[m] = np.where(t_0[m] < t_1[m], t_0[m], t_1[m])
        else:
            distance[m] = np.where(t_0[m] > t_1[m], t_0[m], t_1[m])

        return distance, m

    def intersect_normal(self, xloc, mask):
        m = mask
        norm = np.full(xloc.shape, np.nan, dtype=np.float64)
        norm[m] = xm.normalize(self.param['center'] - xloc[m])
        return norm




# -*- coding:utf-8 -*-
"""
.. Authors
    Novimir pablant <npablant@pppl.gov>

A simple image viewer for use with pixelated detectors.
"""
import numpy as np
import logging

from matplotlib import pyplot
from xicsrt.util import mirplot

def view(image, options=None, **kwargs):
    """
    Plot an image along with column and row summation plots.
    """

    if options is None: options = {}
    options.update(kwargs)
    opt = options


    opt.setdefault('size', None)
    opt.setdefault('pixel_size', 0.1)
    opt.setdefault('aspect', 'equal')
    opt.setdefault('coord', 'index')
    opt.setdefault('scale', 1.0)
    opt.setdefault('units', None)
    opt.setdefault('cmap', 'turbo')
    opt.setdefault('xbound', None)
    opt.setdefault('ybound', None)

    if not opt.get('xlabel'):
        opt['xlabel'] = f'x'
        if opt.get('units'):
            opt['xlabel'] += f" [{opt['units']}]"

    if not opt.get('ylabel'):
        opt['ylabel'] = f'y'
        if opt.get('units'):
            opt['ylabel'] += f" [{opt['units']}]"

    gs = {
        'width_ratios':[3, 1],
        'height_ratios':[1, 3],
        'wspace':0.05,
        'hspace':0.05,
    }
    fig, axs = pyplot.subplots(2, 2, gridspec_kw=gs, figsize=(8,8))
    fig.subplots_adjust(top=0.95, bottom=0.15, left=0.15, right=0.95)
    axs[0, 1].set_axis_off()

    axesdict = {
        'image':axs[1, 0],
        'xsum':axs[0, 0],
        'ysum':axs[1, 1],
    }

    fig.axesdict = axesdict

    for ax in axesdict.values():
        ax.label_outer()

    axesdict['xsum'].sharex(axesdict['image'])
    axesdict['ysum'].sharey(axesdict['image'])

    if opt['coord'] == 'index':
        # Index coordinates.
        # Coordinate system is defined by the bottom left corner of the image.
        # Pixels location defined by bottom left corner.
        x = np.arange(image.shape[0], dtype=np.float64)
        y = np.arange(image.shape[1], dtype=np.float64)
        extent = (0, image.shape[0], 0, image.shape[1])
    elif opt['coord'] == 'cindex':
        # Index coordinates.
        # Coordinate system is defined by the center of the image.
        # Pixels location defined by bottom left corner.
        raise NotImplementedError()
    elif opt['coord'] == 'pixel':
        # Pixel coordinates.
        # Coordinate system is defined by the bottom left corner of the image.
        # Pixels location defined by center.
        x = np.arange(image.shape[0], dtype=np.float64)+0.5
        y = np.arange(image.shape[1], dtype=np.float64)+0.5
        extent = np.array((-0.5, image.shape[0]-0.5, 0.5, image.shape[1]-0.5))
    elif opt['coord'] == 'cpixel':
        # Centered Pixel coordinates.
        # Coordinate system is defined by the center of the image.
        # Pixels location defined by center.
        raise NotImplementedError()
    elif opt['coord'] == 'space':
        # Realspace coordinates.
        # Coordinate system is defined by the center of the image.
        # Pixels location defined by center.
        x = np.linspace(-0.5*opt['size'][0], 0.5*opt['size'][0], image.shape[0])
        y = np.linspace(-0.5*opt['size'][1], 0.5*opt['size'][1], image.shape[1])
        extent = (-0.5*opt['size'][0], 0.5*opt['size'][0], -0.5*opt['size'][1], 0.5*opt['size'][1])
    else:
        raise Exception(f"coord type {opt['coord']} unknown.")

    extent = np.asarray(extent, dtype=float)
    extent *= opt['scale']
    x *= opt['scale']
    y *= opt['scale']

    xsum = np.sum(image, axis=1)
    ysum = np.sum(image, axis=0)

    plotlist = []
    plotlist.append({
        'axes':'image',
        'type':'image',
        'z':np.rot90(image),
        'extent':extent,
        'cmap':opt['cmap'],
        'xlabel':opt['xlabel'],
        'ylabel':opt['ylabel'],
        'xbound':opt['xbound'],
        'ybound':opt['ybound'],
        })
    plotlist.append({
        'axes':'xsum',
        'type':'hline',
        'y':[0.0],
        'color':'black',
        'alpha':0.2,
        'linestyle':'--',
        })
    plotlist.append({
        'axes':'xsum',
        'x':x,
        'y':xsum,
        })
    plotlist.append({
        'axes':'ysum',
        'type':'vline',
        'x':[0.0],
        'color':'black',
        'alpha':0.2,
        'linestyle':'--',
        })
    plotlist.append({
        'axes':'ysum',
        'x':ysum,
        'y':y,
        'ybound':None,
        })
    mirplot.plot_to_axes(plotlist, axesdict)

    if opt.get('aspect','equal') == 'equal':
        _update_lim_aspect(axesdict['image'])
        axesdict['image'].callbacks.connect('ylim_changed', _on_ylims_change)

    return fig

def add_controls(fig):
    """
    Add a set of controls for viewing the image.

    Programming Notes
    -----------------
    Jupyter apparently won't make these controls interactive unless it can see
    them in the local scope. As long as this function is called directly in
    the notebook then everything appears to work ok.
    """
    from matplotlib.widgets import RangeSlider

    im = fig.axesdict['image'].get_images()[0]
    image = im.get_array()

    fig.subplots_adjust(top=0.95, bottom=0.125, left=0.125, right=0.95)

    w = fig.axesdict['image'].get_position().width
    box = [0.125, 0.025, w, 0.02]

    # Create a background based on a count histogram.
    ax = fig.add_axes(box)
    ax.set_axis_off()
    hist, bins = np.histogram(
        image.flatten(),
        range=[image.min() - 0.5, image.max() + 0.5],
        bins=int(image.max() + 1),
        )
    bins_c = (bins[0:-1] + bins[1:]) / 2
    hist[0] = hist[1]
    z = np.reshape(hist, (1,len(hist)))
    ax.imshow(z, aspect='auto', cmap='gist_heat_r')
    ax.set_navigate(False)

    # Create the slider.
    ax = fig.add_axes(box)
    ax.patch.set_alpha(0.0)
    slider = RangeSlider(
        ax,
        '',
        im.norm.vmin,
        im.norm.vmax,
        valinit=(im.norm.vmin, im.norm.vmax),
        alpha=0.5,
        )

    def update(val):
        # The val passed to a callback by the RangeSlider will
        # be a tuple of (min, max)

        # Update the image's colormap
        im.norm.vmin = val[0]
        im.norm.vmax = val[1]

        # Redraw the figure to ensure it updates
        fig.canvas.draw_idle()

    slider.on_changed(update)

    return slider


def _update_lim_aspect(ax):
    """
    Update the data limits (xlim & ylim) to produce an equal aspect given a
    fixed plot size.
    """
    xlim = ax.get_xlim()
    xspan = xlim[1] - xlim[0]
    ylim = ax.get_ylim()
    yspan = ylim[1] - ylim[0]
    h = ax.get_window_extent().height
    w = ax.get_window_extent().width
    w_aspect = h / w
    d_aspect = yspan / xspan

    if d_aspect < w_aspect:
        ylim_new = np.array([-0.5, 0.5]) * xspan * w_aspect + sum(ylim) / 2
        ax.set_ylim(ylim_new)
    else:
        xlim_new = np.array([-0.5, 0.5]) * yspan / w_aspect + sum(xlim) / 2
        ax.set_xlim(xlim_new)


def _on_ylims_change(event_ax):
    """
    An Axes callback to update the data limits after a change in the the data
    limits. This is primarily meant to allow retention af a fixed aspect
    after the user has zoomed into a region.
    """

    # We need to disconnect the callback before calling set_ylim
    # Otherwise we will end up in an infinite loop.
    cid_list = list(event_ax.callbacks.callbacks['ylim_changed'].keys())
    for cid in cid_list:
        event_ax.callbacks.disconnect(cid)

    _update_lim_aspect(event_ax)

    event_ax.callbacks.connect('ylim_changed', _on_ylims_change)



# -*- coding: utf-8 -*-
"""
.. Authors:
   Novimir Antoniuk Pablant <npablant@pppl.gov>

A command line interface for the XICSRT raytracer.
"""

import sys
import numpy as np
import argparse
import io

from xicsrt.util import mirlogging

from xicsrt import xicsrt_raytrace
from xicsrt import xicsrt_multiprocessing
from xicsrt import xicsrt_io

from xicsrt._version import __version__

m_log = mirlogging.getLogger(__name__)

def _get_parser():
    parser = argparse.ArgumentParser(
        "\n  xicsrt"
        ,description = f"""
xicsrt version {__version__}

description:
  Perform an XICSRT raytrace from the command line.

  The input to this command should be an XICSRT configuration dictionary
  in json format. (Pickle and hdf5 formats are also supported.)


example 1:
  xicsrt config.json
  
example 2: 
  python -m xicsrt config.json

"""
        ,formatter_class=lambda prog: argparse.RawDescriptionHelpFormatter(prog, width=79))

    parser.add_argument(
        'config_file',
        type=str,
        nargs='?',
        default='config.json',
        help='The path to the configuration file for this run.',
        )

    parser.add_argument(
        '--numruns',
        type=int,
        default=None,
        metavar='N',
        help="Number of runs.",
        )

    parser.add_argument(
        '--numiter',
        type=int,
        default=None,
        metavar='N',
        help="Number of iterations per run.",
        )

    parser.add_argument(
        '--seed',
        type=int,
        default=None,
        metavar='N',
        help="The random seed to use.",
        )

    parser.add_argument(
        '--save',
        action='store_true',
        help="Save the results.",
        )

    parser.add_argument(
        '--images',
        action='store_true',
        help="Save intersection images.",
        )

    parser.add_argument(
        '--suffix',
        type=str,
        default=None,
        metavar='STR',
        help="A suffix to add to the output files.",
        )

    parser.add_argument(
        '--path',
        type=str,
        default=None,
        metavar='STR',
        help="Directory in which to store output.",
        )

    parser.add_argument(
        '--multiprocessing',
        '--mp',
        action='store_true',
        help="Use multiprocessing.",
        )

    parser.add_argument(
        '--processes',
        type=int,
        default=None,
        metavar='N',
        help="Number of processes to use for muliprocessing.",
        )

    parser.add_argument(
        '--version',
        action='store_true',
        help="Show the version number.",
        )

    parser.add_argument(
        '--debug',
        action='store_true',
        help="Show debugging output in the log.",
        )

    return parser

def run():
    """
    Parse command line arguments and run XICSRT.
    """
    global args

    parser = _get_parser()
    args = parser.parse_args()

    if args.version:
        print(f'{__version__}')
        return

    if args.debug:
        log_level = mirlogging.DEBUG
    else:
        log_level = mirlogging.INFO
    mirlogging.defaultConfig(level=log_level, force=True)

    config = xicsrt_io.load_config(args.config_file)

    if args.suffix:
        config['general']['output_suffix'] = args.suffix
    if args.numruns:
        config['general']['number_of_runs'] = args.numruns
    if args.numiter:
        config['general']['number_of_iter'] = args.numiter
    if args.seed:
        config['general']['random_seed'] = args.seed
    if args.path:
        config['general']['output_path'] = args.path
    if args.save:
        config['general']['save_results'] = args.save
    if args.images:
        config['general']['save_images'] = args.images

    if args.multiprocessing:

        result = xicsrt_multiprocessing.raytrace(config, processes=args.processes)
    else:
        result = xicsrt_raytrace.raytrace(config)

# Generate the module docstring from the helpstring.
parser = _get_parser()
with io.StringIO() as ff:
    parser.print_help(ff)
    help_string = ff.getvalue()
help_string = '| '+'\n| '.join(help_string.splitlines())
__doc__ += '\n'
__doc__ += help_string

if __name__ == "__main__":
    run()



"""
"""

# -*- coding: utf-8 -*-
"""
Created on Thu Jan  3 11:14:52 2019

@author: NKrvavica

This contains vectorized quadratic, cubic and quartic solvers

This code is adapted from https://github.com/NKrvavica/fqs commit d95d684
MIT License - Copyright (c) 2019 NKrvavica

Some modifications have been made to the Api, but the implementation is
identical to fqs.
"""

import math
import numpy as np

def multi_quadratic(a0, b0, c0):
    ''' Analytical solver for multiple quadratic equations
    (2nd order polynomial), based on `numpy` functions.

    Parameters
    ----------
    a0, b0, c0: array_like
        Input data are coefficients of the Quadratic polynomial::

            a0*x^2 + b0*x + c0 = 0

    Returns
    -------
    r1, r2: ndarray
        Output data is an array of two roots of given polynomials.
    '''
    ''' Reduce the quadratic equation to to form:
        x^2 + ax + b = 0'''
    a, b = b0 / a0, c0 / a0

    # Some repating variables
    a0 = -0.5*a
    delta = a0*a0 - b
    sqrt_delta = np.sqrt(delta + 0j)

    # Roots
    r1 = a0 - sqrt_delta
    r2 = a0 + sqrt_delta

    return r1, r2


def multi_cubic(a0, b0, c0, d0, all_roots=True):
    ''' Analytical closed-form solver for multiple cubic equations
    (3rd order polynomial), based on `numpy` functions.

    Parameters
    ----------
    a0, b0, c0, d0: array_like
        Input data are coefficients of the Cubic polynomial::

            a0*x^3 + b0*x^2 + c0*x + d0 = 0

    all_roots: bool, optional
        If set to `True` (default) all three roots are computed and returned.
        If set to `False` only one (real) root is computed and returned.

    Returns
    -------
    roots: ndarray
        Output data is an array of three roots of given polynomials of size
        (3, M) if `all_roots=True`, and an array of one root of size (M,)
        if `all_roots=False`.
    '''

    ''' Reduce the cubic equation to to form:
        x^3 + a*x^2 + bx + c = 0'''
    a, b, c = b0 / a0, c0 / a0, d0 / a0

    # Some repeating constants and variables
    third = 1./3.
    a13 = a*third
    a2 = a13*a13
    sqr3 = math.sqrt(3)

    # Additional intermediate variables
    f = third*b - a2
    g = a13 * (2*a2 - b) + c
    h = 0.25*g*g + f*f*f

    # Masks for different combinations of roots
    m1 = (f == 0) & (g == 0) & (h == 0)     # roots are real and equal
    m2 = (~m1) & (h <= 0)                   # roots are real and distinct
    m3 = (~m1) & (~m2)                      # one real root and two complex

    def cubic_root(x):
        ''' Compute cubic root of a number while maintaining its sign
        '''
        root = np.zeros_like(x)
        positive = (x >= 0)
        negative = ~positive
        root[positive] = x[positive]**third
        root[negative] = -(-x[negative])**third
        return root

    def roots_all_real_equal(c):
        ''' Compute cubic roots if all roots are real and equal
        '''
        r1 = -cubic_root(c)
        if all_roots:
            return r1, r1, r1
        else:
            return r1

    def roots_all_real_distinct(a13, f, g, h):
        ''' Compute cubic roots if all roots are real and distinct
        '''
        j = np.sqrt(-f)
        k = np.arccos(-0.5*g / (j*j*j))
        m = np.cos(third*k)
        r1 = 2*j*m - a13
        if all_roots:
            n = sqr3 * np.sin(third*k)
            r2 = -j * (m + n) - a13
            r3 = -j * (m - n) - a13
            return r1, r2, r3
        else:
            return r1

    def roots_one_real(a13, g, h):
        ''' Compute cubic roots if one root is real and other two are complex
        '''
        sqrt_h = np.sqrt(h)
        S = cubic_root(-0.5*g + sqrt_h)
        U = cubic_root(-0.5*g - sqrt_h)
        S_plus_U = S + U
        r1 = S_plus_U - a13
        if all_roots:
            S_minus_U = S - U
            r2 = -0.5*S_plus_U - a13 + S_minus_U*sqr3*0.5j
            r3 = -0.5*S_plus_U - a13 - S_minus_U*sqr3*0.5j
            return r1, r2, r3
        else:
            return r1

    # Compute roots
    if all_roots:
        roots = np.zeros((3, len(a))).astype(complex)
        roots[:, m1] = roots_all_real_equal(c[m1])
        roots[:, m2] = roots_all_real_distinct(a13[m2], f[m2], g[m2], h[m2])
        roots[:, m3] = roots_one_real(a13[m3], g[m3], h[m3])
    else:
        roots = np.zeros(len(a))  # .astype(complex)
        roots[m1] = roots_all_real_equal(c[m1])
        roots[m2] = roots_all_real_distinct(a13[m2], f[m2], g[m2], h[m2])
        roots[m3] = roots_one_real(a13[m3], g[m3], h[m3])

    return roots


def multi_quartic(a0, b0, c0, d0, e0):
    ''' Analytical closed-form solver for multiple quartic equations
    (4th order polynomial), based on `numpy` functions. Calls
    `multi_cubic` and `multi_quadratic`.

    Parameters
    ----------
    a0, b0, c0, d0, e0: array_like
        Input data are coefficients of the Quartic polynomial::

            a0*x^4 + b0*x^3 + c0*x^2 + d0*x + e0 = 0

    Returns
    -------
    r1, r2, r3, r4: ndarray
        Output data is an array of four roots of given polynomials.
    '''

    ''' Reduce the quartic equation to to form:
        x^4 ax^3 + bx^2 + cx + d = 0'''
    a, b, c, d = b0/a0, c0/a0, d0/a0, e0/a0

    # Some repeating variables
    a0 = 0.25*a
    a02 = a0*a0

    # Coefficients of subsidiary cubic euqtion
    p = 3*a02 - 0.5*b
    q = a*a02 - b*a0 + 0.5*c
    r = 3*a02*a02 - b*a02 + c*a0 - d

    # One root of the cubic equation
    z0 = multi_cubic(1, p, r, p*r - 0.5*q*q, all_roots=False)

    # Additional variables
    s = np.sqrt(2*p + 2*z0.real + 0j)
    t = np.zeros_like(s)
    mask = (s == 0)
    t[mask] = z0[mask]*z0[mask] + r[mask]
    t[~mask] = -q[~mask] / s[~mask]

    # Compute roots by quadratic equations
    r0, r1 = multi_quadratic(1, s, z0 + t) - a0
    r2, r3 = multi_quadratic(1, -s, z0 - t) - a0

    return r0, r1, r2, r3


def eig_quadratic(p):
    '''Finds quadratic roots via numerical eigenvalue solver
    `npumpy.linalg.eigvals` from a 3x3 companion matrix
    '''
    a, b = (p[:, 1]/p[:, 0], p[:, 2]/p[:, 0])
    N = len(a)
    A = np.zeros((N, 2, 2))
    A[:, 1, 0] = 1
    A[:, :, 1] = - np.array([b, a]).T
    roots = np.linalg.eigvals(A)
    return roots


def eig_cubic(p):
    '''Finds cubic roots via numerical eigenvalue solver
    `npumpy.linalg.eigvals` from a 3x3 companion matrix'''
    a, b, c = (p[:, 1]/p[:, 0], p[:, 2]/p[:, 0], p[:, 3]/p[:, 0],)
    N = len(a)
    A = np.zeros((N, 3, 3))
    A[:, 1:, :2] = np.eye(2)
    A[:, :, 2] = - np.array([c, b, a]).T
    roots = np.linalg.eigvals(A)
    return roots


def eig_quartic(p):
    '''Finds quartic roots via numerical eigenvalue solver
    `npumpy.linalg.eigvals` from a 4x4 companion matrix
    '''
    a, b, c, d = (p[:, 1]/p[:, 0], p[:, 2]/p[:, 0],
                  p[:, 3]/p[:, 0], p[:, 4]/p[:, 0])
    N = len(a)
    A = np.zeros((N, 4, 4))
    A[:, 1:, :3] = np.eye(3)
    A[:, :, 3] = - np.array([d, c, b, a]).T
    roots = np.linalg.eigvals(A)
    return roots


def quadratic_roots(p, eigenvalue=False):
    '''
    A caller function for a fast quadratic root solver (2rd order polynomial).

    Parameters
    ----------
    p: array_like
        Input data are coefficients of the Cubic polynomial of the form:

            p[0]*x^2 + p[1]*x + p[2] = 0

        Stacked arrays of coefficient are allowed, which means that ``p`` may
        have size ``(3,)`` or ``(M, 3)``, where ``M>0`` is the
        number of polynomials. Note that the first axis should be used for
        stacking.

    Keywords
    --------
    eigenvalue: bool (False)
        Use an eigenvalue solver based on numpy.linalg.eigvals rather than the
        algebraic method.

    Returns
    -------
    roots: ndarray
        Output data is an array of two roots of given polynomials,
        of size ``(M, 2)``.

    Examples
    --------
    > roots = quadratic_roots([1, 3, 2])
    > roots
    array([[-2.+0.j, -1.+0.j]])

    > roots = quadratic_roots([3, 2, 1])
    > roots
    array([[-0.33333333-0.47140452j, -0.33333333+0.47140452j]])

    > roots = quadratic_roots([[1, 2, 3],
                               [3, 2, 1]])
    > roots
    array([[-2.        +0.j        , -1.        +0.j        ],
           [-0.33333333-0.47140452j, -0.33333333+0.47140452j]])
    '''
    # Convert input to array (if input is a list or tuple)
    p = np.asarray(p)
    if eigenvalue:
        return eig_quadratic(p)

    # If only one set of coefficients is given, add axis
    if p.ndim < 2:
        p = p[np.newaxis, :]

    # Check if four coefficients are given
    if p.shape[1] != 3:
        raise ValueError('Expected 2rd order polynomial with 3 '
                         'coefficients, got {:d}.'.format(p.shape[1]))

    roots = multi_quadratic(*p.T)
    return np.array(roots).T


def cubic_roots(p, eigenvalue=False):
    '''
    A caller function for a fast cubic root solver (3rd order polynomial).

    Equations are based on a closed-form analytical solutions by Cardano.

    Parameters
    ----------
    p: array_like
        Input data are coefficients of the Cubic polynomial of the form:

            p[0]*x^3 + p[1]*x^2 + p[2]*x + p[3] = 0

        Stacked arrays of coefficient are allowed, which means that ``p`` may
        have size ``(4,)`` or ``(M, 4)``, where ``M>0`` is the
        number of polynomials. Note that the first axis should be used for
        stacking.

    Keywords
    --------
    eigenvalue: bool (False)
        Use an eigenvalue solver based on numpy.linalg.eigvals rather than the
        algebraic method.

    Returns
    -------
    roots: ndarray
        Output data is an array of three roots of given polynomials,
        of size ``(M, 3)``.

    Examples
    --------
    > roots = cubic_roots([1, 7, -806, -1050])
    > roots
    array([[ 25.80760451+0.j, -31.51667909+0.j,  -1.29092543+0.j]])

    > roots = cubic_roots([1, 2, 3, 4])
    > roots
    array([[-1.65062919+0.j        , -0.1746854 +1.54686889j,
            -0.1746854 -1.54686889j]])

    > roots = cubic_roots([[1, 2, 3, 4],
                               [1, 7, -806, -1050]])
    > roots
    array([[ -1.65062919+0.j        ,  -0.1746854 +1.54686889j,
             -0.1746854 -1.54686889j],
           [ 25.80760451+0.j        , -31.51667909+0.j        ,
             -1.29092543+0.j        ]])
    '''
    # Convert input to array (if input is a list or tuple)
    p = np.asarray(p)
    if eigenvalue:
        return eig_cubic(p)

    # If only one set of coefficients is given, add axis
    if p.ndim < 2:
        p = p[np.newaxis, :]

    # Check if four coefficients are given
    if p.shape[1] != 4:
        raise ValueError('Expected 3rd order polynomial with 4 '
                         'coefficients, got {:d}.'.format(p.shape[1]))

    roots = multi_cubic(*p.T)
    return np.array(roots).T


def quartic_roots(p, eigenvalue=False):
    '''
    A caller function for a fast quartic root solver (4th order polynomial).

    Equations are based on a closed-form analytical solutions by Ferrari
    and Cardano.

    Parameters
    ----------
    p: array_like
        Input data are coefficients of the Quartic polynomial of the form:

            p[0]*x^4 + p[1]*x^3 + p[2]*x^2 + p[3]*x + p[4] = 0

        Stacked arrays of coefficient are allowed, which means that ``p`` may
        have size ``(5,)`` or ``(M, 5)``, where ``M>0`` is the
        number of polynomials. Note that the first axis should be used for
        stacking.

    Keywords
    --------
    eigenvalue: bool (False)
        Use an eigenvalue solver based on numpy.linalg.eigvals rather than the
        algebraic method.

    Returns
    -------
    roots: ndarray
        Output data is an array of four roots of given polynomials,
        of size ``(M, 4)``.

    Examples
    --------
    > roots = quartic_roots([1, 7, -806, -1050, 38322])
    > roots
    array([[-30.76994812-0.j,  -7.60101564+0.j,   6.61999319+0.j,
             24.75097057-0.j]])

    > roots = quartic_roots([1, 2, 3, 4, 5])
    > roots
    array([[-1.28781548-0.85789676j, -1.28781548+0.85789676j,
             0.28781548+1.41609308j,  0.28781548-1.41609308j]])

    > roots = quartic_roots([[1, 2, 3, 4, 5],
                               [1, 7, -806, -1050, 38322]])
    > roots
    array([[ -1.28781548-0.85789676j,  -1.28781548+0.85789676j,
              0.28781548+1.41609308j,   0.28781548-1.41609308j],
           [-30.76994812-0.j        ,  -7.60101564+0.j        ,
              6.61999319+0.j        ,  24.75097057-0.j        ]])
    '''
    # Convert input to an array (if input is a list or tuple)
    p = np.asarray(p)
    if eigenvalue:
        return eig_quartic(p)

    # If only one set of coefficients is given, add axis
    if p.ndim < 2:
        p = p[np.newaxis, :]

    # Check if all five coefficients are given
    if p.shape[1] != 5:
        raise ValueError('Expected 4th order polynomial with 5 '
                         'coefficients, got {:d}.'.format(p.shape[1]))

    roots = multi_quartic(*p.T)
    return np.array(roots).T


# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

A set of physcis utilities and vector convenience functions for XICSRT.
"""

import numpy as np

from scipy import constants as const

def energy_from_wavelenth(wave):
    energy = const.h * const.c / wave / const.e * 1e10
    return energy

def wavelength_from_energy(energy):
    wave = const.h * const.c / energy / const.e * 1e10
    return wave




# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir pablant <npablant@pppl.gov>

Define the :class:`ShapeObject` class.
"""

import numpy as np
from copy import deepcopy

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._TraceObject import TraceObject
from xicsrt.tools import xicsrt_math as xm


@dochelper
class ShapeObject(TraceObject):
    """
    The base class for intersections of rays with surfaces in XICSRT.

    This base class should be used to define intersections with various shapes
    such as planes, spheres and toroids.
    """

    def intersect(self, rays):
        """
        Calculate the location and normal of the surface at the ray
        intersections.

        Specific shape objects can reimplement this method, or alternatively
        reimplement the :meth:`intersect_location` and :meth:`intersect_normal`
        methods.

        Programming Notes
        -----------------
        Currently the expectation is that intersect has made copies of
        ray['origin'] and ray['mask'] before any calculations. This is done for
        two reasons: 1. provide more information for the interactions. 2. it is
        much easier to read and understand the code this way. From a memory
        efficiency standpoint it would be better to modify these arrays in place
        instead.
        """
        xloc, mask = self.intersect_location(rays)
        norm = self.intersect_normal(rays, xloc, mask)

        return xloc, norm, mask

    def intersect_location(self, rays):
        """
        Calculate the surface location at the ray intersections.

        This base-class just returns a copy of the ray origin.
        """
        xloc = rays['origin'].copy()
        mask = rays['mask'].copy()
        return xloc, mask

    def intersect_normal(self, xloc, mask):
        """
        Calculate the surface normal at the ray intersection locations.

        Normals are not defined for this base-class; an array of np.nan will
        always be returned.
        """
        norm = np.full(xloc.shape, np.nan, dtype=np.float64)
        return norm

    def location_from_distance(self, rays, dist, mask=None):
        """
        Calculate 3D locations given a distance along the rays.
        """
        if mask is None: mask = rays['mask']
        O = rays['origin']
        D = rays['direction']
        m = mask

        X = np.full(O.shape, np.nan, dtype=np.float64)
        X[m] = O[m] + D[m] * dist[m, np.newaxis]

        return X


# -*- coding: utf-8 -*-
"""
Authors
-------
  - Novimir A. Pablant <nablant@pppl.gov>
  - Yevgeniy Yakusevich <eugenethree@gmail.com>
"""
import logging
import numpy as np
from copy import copy

from xicsrt.util import profiler
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.sources._XicsrtPlasmaGeneric import XicsrtPlasmaGeneric

import xicsrt.tools.xicsrt_math as xm

@dochelper
class XicsrtPlasmaToroidal(XicsrtPlasmaGeneric):
    """
    A plasma object with toroidal geometry and a circular cross-section.
    """
        
    def default_config(self):
        config = super().default_config()
        config['major_radius'] = 0.0
        config['minor_radius'] = 0.0
        config['torus_origin'] = np.array([0.0, 0.0, 0.0])
        config['emissivity_scale']  = 1.0
        config['temperature_scale'] = 1.0
        config['velocity_scale']    = 1.0
        return config

    def flx_from_car(self, point_car):
        point_flx = xm.tor_from_car(point_car - self.param['torus_origin'], self.param['major_radius'])
        point_flx[0] = point_flx[0]**2
        point_flx[0] /= self.param['minor_radius']
        return point_flx

    def rho_from_car(self, point_car):
        point_flx = self.flx_from_car(point_car)
        return np.sqrt(point_flx[0])

    def car_from_flx(self, point_flx):
        point_tor = copy(point_flx)
        point_tor[...,0] = np.sqrt(point_tor[...,0])*self.param['minor_radius']
        point_car = xm.car_from_tor(point_tor, self.param['major_radius'])
        return point_car

    def bundle_generate(self, bundle_input):

        profiler.start("Bundle Input Generation")
        m = bundle_input['mask']

        # Attempt to generate the specified number of bundles, but throw out
        # bundles that our outside of the last closed flux surface.
        #
        # This loop was setup for VMEC. Here we could do this as a single
        # vectorized operation instead.
        rho = np.zeros(len(m[m]))
        for ii in range(len(m[m])):
            # convert from cartesian coordinates to normalized radial coordinate.
            profiler.start("Fluxspace from Realspace")
            rho[ii] = self.rho_from_car(bundle_input['origin'][m][ii, :])
            profiler.stop("Fluxspace from Realspace")

        # evaluate emissivity, temperature and velocity at each bundle location.
        bundle_input['temperature'][m] = self.get_temperature(rho) * self.param['temperature_scale']
        bundle_input['emissivity'][m] = self.get_emissivity(rho) * self.param['emissivity_scale']
        bundle_input['velocity'][m] = self.get_velocity(rho) * self.param['velocity_scale']

        fintest = np.isfinite(bundle_input['temperature'])

        m &= fintest

        profiler.stop("Bundle Input Generation")

        return bundle_input



# -*- coding: utf-8 -*-
"""
Authors
-------
- Novimir A. Pablant <npblant@pppl.gov>
"""
import numpy as np
import logging

import copy

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.tools import xicsrt_string

from xicsrt.tools import xicsrt_misc
from xicsrt import xicsrt_config

@dochelper
class ConfigObject():
    """
    A base class for any objects with a configuration.
    """

    def __init__(self, config=None, strict=None, initialize=None):
        if initialize is None: initialize = True

        self.name = self.__class__.__name__
        self.log = logging.getLogger('xicsrt').getChild(self.name)

        self.config = self.default_config()
        self.update_config(config, strict=strict)
        self.check_config()

        self.param = copy.deepcopy(self.config)
        self.param = xicsrt_misc._convert_to_numpy(self.param, inplace=True)

        if initialize:
            self.setup()
            self.check_param()
            self.initialize()

    def default_config(self):
        """
        class_name
          Automatically generated.

        yo_mama
          Is a wonderful person!
        """

        config = dict()
        config['class_name'] = self.__class__.__name__
        config['yo_mama'] = 'Is a beautiful person and she loves you.'
        return config

    def get_config(self):
        return self.config

    def check_config(self):
        """
        Check the config before copying to the internal param. This is called
        during object instantiation (`__init__`) and therefore before `setup` is
        called.
        """
        pass

    def setup(self):
        """
        Perform any setup actions that are needed prior to initialization.
        """
        pass

    def check_param(self):
        """
        Check the internal parameters prior to initialization. This will be
        called after `setup` and before `initialize`.
        """
        pass

    def initialize(self):
        """
        Initialize the object.
        """
        pass

    def update_config(self, config_new, **kwargs):
        """
        Overwrite any config values in this object with the ones given. This
        will be done recursively for all nested dictionaries.
        """
        xicsrt_config.update_config(self.config, config_new, **kwargs)






# -*- coding: utf-8 -*-
"""
.. Authors
     Nathan Bartlett <nbb0011@auburn.edu>
     Novimir Pablant <npablant@pppl.gov>

A set of apertures for ray filtering.
"""

import numpy as np
import xicsrt.tools.xicsrt_math as xm

def aperture_mask(X_local, m, aperture_info):
    """
    Generate a mask array for the given aperture (or array of apertures).
    """

    if aperture_info is None:
        return m

    # Convert the aperture info to an array, if was not already.
    aperture_info = xm.toarray_1d(aperture_info)

    m_out = m.copy()
    for aperture in aperture_info:
        aperture = _aperture_defaults(aperture)

        m_test = m.copy()
        m_test = aperture_selector(X_local, m_test, aperture, _internal=True)

        logic = aperture['logic']
        if logic == 'and':
            m_out[m] &= m_test[m]
        elif logic == 'not':
            m_out[m] &= ~m_test[m]
        elif logic == 'or':
            m_out[m] |= m_test[m]
        elif logic == 'nand':
            m_out[m] = ~(m_out[m] & m_test[m])
        elif logic == 'nor':
            m_out[m] = ~(m_out[m] | m_test[m])
        elif logic == 'xor':
            m_out[m] ^= m_test[m]
        elif logic == 'xnor':
            m_out[m] = ~(m_out[m] ^ m_test[m])
        else:
            raise Exception(f'Aperture logic "{logic}" is not known.')
            
    return m_out


def aperture_selector(X_local, m, aperture, _internal=False):
    """
    Will call the appropriate aperture function for the given aperture name.

    .. Note::
       This selector and all the individual function will modify the mask array
       in place.
    """

    if not _internal:
        aperture = _aperture_defaults(aperture)

    shape = aperture['shape']
    if shape == 'none':
        func = aperture_none
    elif shape == 'circle':
        func = aperture_circle
    elif shape == 'square':
        func = aperture_square
    elif shape == 'rectangle':
        func = aperture_rectangle
    elif shape == 'ellipse':
        func = aperture_ellipse
    elif shape == 'triangle':
        func = aperture_triangle
    else:
        raise Exception(f'Aperture shape: "{shape}" is not implemented.')

    return func(X_local, m, aperture)


def _aperture_defaults(aperture):
    new = {
        'shape':None,
        'origin':None,
        'logic':None,
        }
    new.update(aperture)

    if new['origin'] is None: new['origin'] = np.array([0.0, 0.0])
    if new['shape'] is None: new['shape'] = 'none'
    if new['logic'] is None: new['logic'] = 'and'

    new['origin'] = xm.toarray_1d(new['origin'])
    new['shape'] = new['shape'].lower()
    new['logic'] = new['logic'].lower()

    if 'size' in new:
        new['size'] = xm.toarray_1d(new['size'])

    if 'vertices' in new:
        new['vertices'] = np.asarray(new['vertices'])

    return new


def aperture_none(X_local, m, aperture):
    """
    An empty aperture object.
    """
    return m


def aperture_circle(X_local, m, aperture):
    """
    A circular Aperture.

    name: 'circle'

    size: [radius]
      Contains the radius of the aperture.
    """
    origin_x = aperture['origin'][0]
    origin_y = aperture['origin'][1]
    size = aperture['size'][0]
    m[m] &= (((X_local[m,0] - origin_x)**2 + (X_local[m,1] - origin_y)**2) < size**2)
    
    return m


def aperture_square(X_local, m, aperture):
    """
    A square Aperture.

    name: 'square'

    size: [x, y]
      Contains the x and y size (full width) of the aperture.
    """
    size = aperture['size'][0]
    origin_x = aperture['origin'][0]
    origin_y = aperture['origin'][1]
    m[m] &= (np.abs((X_local[m, 0] - origin_x)) < size/2)
    m[m] &= (np.abs((X_local[m, 1] - origin_y)) < size/2)

    return m


def aperture_rectangle(X_local, m, aperture):
    """
    A rectangular Aperture.

    name: 'rectangle'

    size: [x, y]
      Contains the x and y size (full width) of the aperture.
    """
    size_x = aperture['size'][0]
    size_y = aperture['size'][1]
    origin_x = aperture['origin'][0]
    origin_y = aperture['origin'][1]
    m[m] &= (np.abs((X_local[m,0] - origin_x)) < size_x / 2)
    m[m] &= (np.abs((X_local[m,1] - origin_y)) < size_y / 2)
    
    return m


def aperture_ellipse(X_local, m, aperture):
    """
    An elliptical Aperture.

    name: 'ellipse'

    size: [x, y]
      Contains the x and y size (full width) of the aperture.
    """
    size_x = aperture['size'][0]
    size_y = aperture['size'][1]
    origin_x = aperture['origin'][0]
    origin_y = aperture['origin'][1]
    m[m] &= ((((X_local[m,0] - origin_x)/size_x)**2 + ((X_local[m,1] - origin_y)/size_y)**2) < 1)
    
    return m


def aperture_triangle(X_local, m, aperture):
    """
    An triangular aperture defined by three vertices.

    name: 'ellipse'

    vertices: [[x0, y0], [x1,y1], [x2,y2]]
      Contains the three vertices of the aperture.
    """

    m[m] = xm.point_in_triangle_2d(
        X_local[m,0:2],
        aperture['vertices'][0,0:2]+aperture['origin'][0:2],
        aperture['vertices'][1,0:2]+aperture['origin'][0:2],
        aperture['vertices'][2,0:2]+aperture['origin'][0:2],
        )

    return m

























# -*- coding: utf-8 -*-
"""
..  Authors:
    Novimir Antoniuk Pablant <npablant@pppl.gov>

A simple example consisting only of a point source and a spherical crystal.

Description
-----------

1.
Create a new user configuration dictionary.

The entries that we put into this config will overwrite the defaults
that are defined within xicsrt. The config can potentially contain the
following sections:

- general
- sources
- optics
- filters
- scenario

2.
Create a section that contains the general raytracer configuration.

number_of_iter
  Perform raytracing the given number of times. The output from all
  the iterations will be combined. Performing multiple iterations allows
  a large number of rays to be traced without running into memory limits.
save_images
  If set to true, images will be saved to the output directory (which
  we have not specified in this example.

3.
Create the section that contains the sources.
Then define a source, cleverly named 'source'.

class_name
  The type of source object to create.
intensity
  The number of rays to launch in each iteration.
wavelength
  The nominal wavelength of the source emission.
spread
   The angular spread of the source (in radians).

4.
Create the section that contains the optics.
In this case we only define one optic: a detector.

class_name
  The type of optic object to create.
origin
  The location of this optic.
zaxis
  The direction the optics is pointing. For all of the standard
  optics that come with xicrt, the zaxis is the normal direction.
xsize
  The size of the optic along the xaxis.
  Corresponds to the 'width' of the optic.
ysize
  The size of the optic along the yaxis.
  Corresponds to the 'height' of the optic.

5.
Finally we pass the configuration to the XICSRT raytracer to perform
the actual raytracing. The `results` is a dictionary with the full
trace history along with images at the detector.
"""

import numpy as np
import xicsrt
xicsrt.warn_version('0.8')

# 1.
config = {}

# 2.
config['general'] = {}
config['general']['number_of_iter'] = 5
config['general']['save_images'] = False

# 3.
config['sources'] = {}
config['sources']['source'] = {}
config['sources']['source']['class_name'] = 'XicsrtSourceDirected'
config['sources']['source']['intensity'] = 1e3
config['sources']['source']['wavelength'] = 3.9492
config['sources']['source']['spread'] = np.radians(5.0)

# 4.
config['optics'] = {}
config['optics']['detector'] = {}
config['optics']['detector']['class_name'] = 'XicsrtOpticDetector'
config['optics']['detector']['origin'] = [0.0, 0.0, 1.0]
config['optics']['detector']['zaxis']  = [0.0, 0.0, -1]
config['optics']['detector']['xsize']  = 0.2
config['optics']['detector']['ysize']  = 0.2

# 5.
results = xicsrt.raytrace(config)




# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.sources._XicsrtSourceGeneric import XicsrtSourceGeneric

@dochelper
class XicsrtSourceFocused(XicsrtSourceGeneric):
    """
    An extended rectangular ray source that allows focusing towards a target.

    This is different to a SourceDirected in that the emission cone is aimed
    at the target for every location in the source. The SourceDirected instead
    uses a fixed direction for emission.
    """
    
    def default_config(self):
        """
        target
          The target at which to aim the emission cone at each point in the
          source volume. The emission cone aimed at the target will have
          an angular spread defined by `spread`.
        """
        config = super().default_config()
        config['target'] = None
        return config
    
    def generate_direction(self, origin):
        normal = self.make_normal_focused(origin)
        D = super().random_direction(normal)
        return D
    
    def make_normal_focused(self, origin):
        # Generate ray from the origin to the focus.
        normal = self.param['target'] - origin
        normal = normal / np.linalg.norm(normal, axis=1)[:, np.newaxis]
        return normal



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>
"""


import numpy as np
import logging
import os
import copy
from xicsrt.objects._ConfigObject import ConfigObject
from xicsrt.tools import xicsrt_misc

from xicsrt import _version

log = logging.getLogger(__name__)

try:
    import xicsrt_contrib
except:
    log.debug('The xicrsrt_contrib package is not installed.')
else:
    log.debug('The xicrsrt_contrib package is installed.')


def default_config():
    """
    number_of_iter : int (1)
      Number of raytracing iterations to perform for each raytracing run.
      Iterations are performed in a single process and with a single
      initialization of the raytracing objects and the random seed.
      All iterations will be combined before saving images or output files.

    number_of_runs : int (1)
      Number of raytracing runs to perform. For each run, the specified
      number of iterations will be performed. Raytracing runs can be
      performed in separate processes enabling the use of multiprocessing.
      At the start of each run all raytracing objects will be created and
      initialized. Images and output files will be saved after each run.

    random_seed : int (None)
      A random seed to initialize the pseudo-random number generator. If
      random_seed is equal to None then a random seed will be provided by
      the python/numpy internals. When a integer seed is provided and
      multiple raytracing runs are performed, the random seed will be
      incremented by one for each successive run. This seed can be used to
      make raytracing runs reproducible on the level of individual rays;
      however, reproducibility is not guaranteed between different versions
      of XICSRT or different versions of Python. Random seed initialization
      performed using `np.random.seed()`.

    pathlist : list (list())
      A list of paths that contain user defined raytracing modules (sources,
      optics, filters, apertures, etc.). These paths will be searched for
      filenames that match the requested 'class_name' in the object config.
      User defined paths are searched before the builtin or contrib paths.

    pathlist_default : list
      A list of paths to the builtin and contributed raytracing objects.
      This option should not be changed by the user, but can be useful
      for inspection to see what directories are actually being searched.

    strict_config_check : bool (True)
      Use strict checking to ensure that the config dictionary only contains
      valid options for the given object. This helps avoid unexpected
      behavior as well and alerting to typos in the configuration. When set
      to `False`, unrecognized config options will be quietly ignored.

    output_path : str (None)
      Path in which to save images and results from the raytracing run. If
      note then the current working path will be used. Use the option
      `make_directories` to control directory creation.

    output_prefix : str ("xicsrt")
      Filenames for images and results are automatically generated. Use
      this option to add a prefix to the beginning of all filenames. An
      underscore will be automatically added after the prefix. For images
      the following format will be used: "prefix_optic_suffix_run.tif".
      For example: "xicsrt_detector_scan01_0000.tif".

    output_suffix : str (None)
      If present this string will be added to automatically generated
      filenames after the optic name and before the run_suffix. See the
      option `output_prefix` for more information.

    output_run_suffix : str (None)
      This option is used internally and should not be set by the user.

    image_ext : str ('.tif')
      Controls the file format of the saved images. Any format supported by
      the `pillow` package can be used; .tif images are recommended.

    results_ext : str ('.hdf5')
      Controls the file format for saving the results dictionary. Currently
      hdf5 (.hdf5, .h5), pickle (.pickle, .pkl) and json (.json) file
      formats are supported. The json format is not recommended as it may
      lead to very large file sizes!

    make_directories : bool (False)
      Controls whether the output path should be created if it does not
      already exist. If False an error will be raised if the output path
      does not exist.

    keep_meta : bool (True)
      Controls whether to calculate and keep metadata and statistics
      relating to the raytracing.

    keep_images : bool (True)
      Controls whether to generate and keep pixelated images of the ray
      intersections at each optic.  Control of image generation for
      individual optics can also be set within the object specific
      config sections.

    keep_history : bool (True)
      Controls whether to calculate and keep the raytracing history. Rays
      will be sorted into 'lost' and 'found' rays, where found rays are
      those that reach the final optic element. The 'found' ray history
      will be kept in full, the 'lost' ray history will be truncated (see
      option `history_max_lost`).

      The ray history provides a great deal of information about the
      raytracing and enables ray plotting (2d and 3d) and post processing.
      However, turning on the ray history also *greatly* increases memory
      usage since the rays must be duplicated and saved for every optical
      element. If only final intersection images are required, consider
      setting this option to `False` to improve raytracing performance.

    history_max_lost : int (10000)
      Number of 'lost' rays to retain in the raytrace history. Lost rays
      are those that are launched from the source but that do not reach the
      last optical element (typically a detector). For many x-ray raytracing
      applications the number of lost rays will be very large, and retention
      of all lost rays would quickly exhaust available system memory. To
      avoid memory issues, while still retaining some lost rays for
      diagnostic purposes, a randomized truncation of the lost rays is
      performed.

    save_config : bool (False)
      Option whether or not to save the config dictionary. Output format
      is currently limited to json format (hdf5 and pickle coming soon).

    save_images : bool (False)
      Controls saving of images. Images will be saved for every run, and a
      combined image will be saved at the conclusion of all runs. Control of
      output for individual optics can be set within the object specific
      config sections. Image format will be determined by the option
      `image_ext` (default .tif). Images will be saved to the `output_path`.

    save_results : bool (False)
      Controls saving of the raytracing results dictionary. The contents of
      the results dictionary are controlled by `keep_meta`, `keep_images`
      and `keep_history`. Results will only be saved after all runs are
      completed. Output format will be determined by the option
      `results_ext` (default .hdf5). OUtput will be saved to the
      `output_path`.

    print_results : bool (True)
      Control text output to terminal of raytracing summary and optics
      specific information. (Note: control of logging/debugging output is
      controlled through a separate option that is not yet implemented.)

    version: string
      The version number of xicsrt when this config was created.
      This option is set internally and should not be modified by the user.
    """
    config = dict()
    config['general'] = dict()
    config['general']['version'] = _version.__version__
    config['general']['number_of_iter'] = 1
    config['general']['number_of_runs'] = 1
    config['general']['random_seed'] = None
    config['general']['pathlist'] = []
    config['general']['pathlist_default'] = get_pathlist_default()
    config['general']['strict_config_check'] = True

    config['general']['output_path'] = None
    config['general']['output_prefix'] = 'xicsrt'
    config['general']['output_suffix'] = None
    config['general']['output_run_suffix'] = None
    config['general']['image_ext'] = '.tif'
    config['general']['results_ext'] = '.hdf5'
    config['general']['config_ext'] = '.json'
    config['general']['make_directories'] = False

    config['general']['keep_meta'] = True
    config['general']['keep_images'] = True
    config['general']['keep_history'] = True

    config['general']['history_max_lost'] = 10000

    config['general']['save_config'] = False
    config['general']['save_images'] = False
    config['general']['save_results'] = False

    config['general']['print_results'] = True

    config['sources'] = dict()
    config['optics'] = dict()
    config['filters'] = dict()
    config['scenario'] = dict()

    return config


def get_config(config_user=None):
    config = default_config()
    update_config(config, config_user, strict=False, update=True)

    return config


def refresh_config(config_new):
    """
    When a config file is loaded from a another system or from a different user
    it may contain default values that are not appropriate for the current
    environment. This function will overwrite these options with new default
    values where appropriate.
    """

    refresh = {}
    refresh['general'] = {}
    refresh['general']['pathlist_default'] = None

    config_new = copy.deepcopy(config_new)
    update_config(
        config_new,
        refresh,
        strict=False,
        update=True,
        ignore_none=False,
        )

    config = default_config()
    update_config(
        config,
        config_new,
        strict=False,
        update=True,
        ignore_none=True,
        )

    return config


def get_pathlist_default():
    """
    Return a list of the default sources and optics directories.
    These locations will be based on the location of this module.
    """

    pathlist = []
    pathlist = _add_pathlist_builtin(pathlist)
    pathlist = _add_pathlist_contrib(pathlist)
    return pathlist


def _add_pathlist_builtin(pathlist):
    # Add paths to built-in objects.
    path_module = os.path.dirname(os.path.abspath(__file__))
    pathlist.append(os.path.join(path_module, 'filters'))
    pathlist.append(os.path.join(path_module, 'sources'))
    pathlist.append(os.path.join(path_module, 'optics'))

    return pathlist


def _add_pathlist_contrib(pathlist):
    # Check if the xicsrt_contrib module was successfully imported.
    if not 'xicsrt_contrib' in globals():
        return pathlist

    # Add paths to the xicsrt_contrib objects.
    path_module = os.path.dirname(os.path.abspath(xicsrt_contrib.__file__))
    pathlist.append(os.path.join(path_module, 'filters'))
    pathlist.append(os.path.join(path_module, 'sources'))
    pathlist.append(os.path.join(path_module, 'optics'))

    return pathlist


def config_to_numpy(obj):
    xicsrt_misc._convert_to_numpy(obj)
    return obj


def config_from_numpy(obj):
    xicsrt_misc._convert_from_numpy(obj)
    return obj


def update_config(
        config,
        config_new,
        strict=None,
        update=None,
        ignore_none=None,
        ):
    """
    Overwrite any values in the given config dict with the values in the
    config_new dict.  This will be done recursively to allow nested
    dictionaries.

    keywords:
      strict (True)
        If True then an error will be raised if an option is found in
        the user dict that is not found in the default dict.

      update (False)
        If True any unmatched options that are found will be retained.
        When False they will simply be ignored. This option has no effect
        unless strict = False.

      ignore_none (False)
        If True any options found in config_new with a value of None will
        be ignored.

    """
    _update_config_dict(config, config_new, strict, update, ignore_none)
    return config


def _update_config_dict(
        config,
        config_new,
        strict=None,
        update=None,
        ignore_none=None,
        ):
    """
    Recursive worker function for `update_config`.
    """
    if strict is None:
        strict = True
    if update is None:
        update = False
    if ignore_none is None:
        ignore_none = False

    if config_new is None:
        return

    for key in config_new:
        if not key in config:
            if strict:
                raise Exception("User option not recognized: {}".format(key))
            if update:
                config[key] = config_new[key]
        else:
            if isinstance(config[key], dict) and isinstance(config_new[key], dict):
                _update_config_dict(
                    config[key],
                    config_new[key],
                    strict=strict,
                    update=update,
                    ignore_none=ignore_none
                    )
            else:
                if ignore_none and config_new[key] is None:
                    pass
                else:
                    config[key] = config_new[key]



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>
"""

import numpy as np   
from scipy.stats import cauchy        
import scipy.constants as const

from xicsrt.util import profiler
from xicsrt.tools import xicsrt_voigt
from xicsrt.tools import xicsrt_spread
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.objects._RayArray import RayArray
from xicsrt.objects._GeometryObject import GeometryObject

@dochelper
class XicsrtSourceGeneric(GeometryObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter_objects = []
            
    def default_config(self):
        """
        xsize
          The size of this element along the xaxis direction.

        ysize
          The size of this element along the yaxis direction.

        zsize
          The size of this element along the zaxis direction.

        spatial_dist : string ('uniform')
          | The type of angular distribution to use for the emitted rays.
          | Available distributions (default is 'uniform'):
          |
          | uniform
          |   Uniform spatial distribution of rays within a rectangular cuboid
          |   defined by xsize, ysize and zsize. The sizes are interpreted as
          |   full widths.
          |
          | gaussian
          |   Gaussian spatial distribution of rays with a fwhm in each
          |   dimension defined by xsize, ysize and zsize. The sizes are
          |   interpreted as full-width-at-half-max (fwhm).

        angular_dist : string ('isotropic')
          | The type of angular distribution to use for the emitted rays.
          | Available distributions (default is 'isotropic'):
          |
          | isotropic
          |   Isotropic emission (uniform spherical) emitted in a cone (circular
          |   cross-section) with a half-angle of 'spread'. The axis of the
          |   emission cone is aligned along the z-axis. 'spread' must be a
          |   single value (scalar or 1-element array).
          | isotropic_xy
          |   Isotropic emission (uniform spherical) emitted in a truncated-cone
          |   (rectangular cross-section) with different x and y half-angles.
          |   'spread' can contain either 1, 2 or 4 values:
          |   s or [s]
          |     A single value that will be used for both the x and y directions.
          |   [x, y]
          |     Two values values that will be used for the x and y directions.
          |   [xmin, xmax, ymin, ymax]
          |     For values that define the asymmetric exent in x and y directions.
          |     Example: [-0.1, 0.1, -0.5, 0.5]
          | flat
          |   Flat emission (uniform planar) emitted in a cone (circular cross-
          |   section) with a half-angle of 'spread'.
          | flat_xy
          |   Flat emission (uniform planar) emitted in a truncated-cone
          |   (rectangular cross-section) with different x and y half-angles.
          |   'spread' can contain either 1, 2 or 4 values, see above.
          | gaussian
          |   Emission with angles away from the z-axis having a Gaussian
          |   distribution (circular cross-section). The 'spread' defines the
          |   Half-Width-at-Half-Max (HWHM) of the distribution. 'spread' must
          |   be a single value (scalar or 1-element array).
          | gaussian_flat
          |   !! Not implemented !!
          |   Cross-section of emission (intersection with constant-z plane) will
          |   have a Gaussian distribution.

        spread : float or array (np.pi) [radians]
          The angular spread for the emission cone. The spread defines the
          half-angle of the emission cone. See 'angular_dist' for detailed
          documentation.

        intensity : int or float
          The number of rays for this source to emit. This should be an
          integer value unless `use_poisson = True`.

          Note: If filters are attached, this will be the number of rays
          emitted before filtering.

        use_poisson : bool (False)
          If `True` the `intenisty` will be treated as the expected value for
          a Poisson distribution and the number of rays will be randomly
          picked from a Poisson distribution. This is setting is typically
          only used internally for Plasma sources.

        wavelength_dist : str ('voigt')
          The type of wavelength distribution for this source.
          Possible values are: 'voigt', 'uniform', 'monochrome'.

          Note: A monochrome distribution can also be achieved by using a
          'voigt' distribution with zero linewidth and temperature.

        wavelength : float (1.0) [angstroms]
          Only used if `wavelength_dist = "monochrome" or "voigt"`
          Central wavelength of the distribution, in Angstroms.

        wavelength_range: tuple [angstroms]
          Only used if `wavelength_dist = "uniform"`
          The wavelength range of the distribution, in Angstroms.
          Must be a 2 element tuple, list or array: (min, max).

        linewidth : float (0.0) [1/s]
          Only used if `wavelength_dist = "voigt"`
          The natural width of the emission line.
          This will control the Lorentzian contribution to the the overall Voigt
          profile. If linewidth == 0, the resulting wavelength distribution will
          be gaussian.

          To convert from a fwhm in [eV]:
          linewidth = 2*pi*e/(h*fwhm_ev)

          To translate from linewidth to gamma in the Voigt equation:
          gamma = linewidth * wavelength**2 / (4*pi*c*1e10)

        mass_number : float (1.0) [au]
          Only used if `wavelength_dist = "voigt"`
          The mass of the emitting atom in atomic units (au). This mass in used
          to convert temperature into line width. See temperature option.

        temperature : float (0.0) [eV]
          Only used if `wavelength_dist = "voigt"`
          The temperature of the emission line.
          This will control the Gaussian contribution to the overall Voigt
          profile. If temperature == 0, the resulting wavelength distribution
          will be Lorentzian.

          To translate from temperature to sigma in the Voigt equation:
          sigma = np.sqrt(temperature/mass_number/amu_kg/c**2*ev_J)*wavelength

        velocity
          No documentation yet. Please help improve XICSRT!

        filters
          No documentation yet. Please help improve XICSRT!

        """
        config = super().default_config()

        config['xsize'] = 0.0
        config['ysize'] = 0.0
        config['zsize'] = 0.0

        config['intensity'] = 0.0
        config['use_poisson'] = False
        config['spatial_dist'] = 'uniform'
        config['angular_dist'] = 'isotropic'
        config['spread'] = np.pi

        # Possible values: 'monochrome', 'voigt', 'uniform
        config['wavelength_dist'] = 'voigt'

        # Only used for wavelength_dist = 'voigt' or 'monochrome'
        config['wavelength'] = 1.0

        # Only used for wavelength_dist = 'voigt'
        config['mass_number'] = 1.0
        config['linewidth'] = 0.0
        config['temperature'] = 0.0
        config['velocity'] = np.array([0.0, 0.0, 0.0])

        # Only used for wavelength_dist = 'uniform'
        config['wavelength_range'] = np.array([0.0, 0.0])
        
        config['filters'] = []

        return config

    def initialize(self):
        super().initialize()
        
        if self.param['use_poisson']:
            self.param['intensity'] = np.random.poisson(self.param['intensity'])
        else:
            if self.param['intensity'] < 1:
                raise ValueError('intensity of less than one encountered. Turn on poisson statistics.')
        self.param['intensity'] = int(self.param['intensity'])
        
    def generate_rays(self):
        rays = RayArray()
        profiler.start('generate_rays')
        
        profiler.start('generate_origin')
        rays['origin'] = self.generate_origin()
        profiler.stop('generate_origin')

        profiler.start('generate_direction')
        rays['direction'] = self.generate_direction(rays['origin'])
        profiler.stop('generate_direction')

        profiler.start('generate_wavelength')
        rays['wavelength'] = self.generate_wavelength(rays['direction'])
        profiler.stop('generate_wavelength')

        profiler.start('generate_weight')
        rays['weight'] = self.generate_weight()
        profiler.stop('generate_weight')
        
        profiler.start('generate_mask')
        rays['mask'] = self.generate_mask()
        profiler.stop('generate_mask')
        
        profiler.start('filter_rays')
        rays = self.ray_filter(rays)
        profiler.stop('filter_rays')        
        
        profiler.stop('generate_rays')
        return rays
     
    def generate_origin(self):

        if self.param['spatial_dist'] == 'uniform':
            # Origins for a uniform distribution of rays
            x_offset = np.random.uniform(-1 * self.param['xsize']/2, self.param['xsize']/2, self.param['intensity'])
            y_offset = np.random.uniform(-1 * self.param['ysize']/2, self.param['ysize']/2, self.param['intensity'])
            z_offset = np.random.uniform(-1 * self.param['zsize']/2, self.param['zsize']/2, self.param['intensity'])

        elif self.param['spatial_dist'] == 'gaussian':
            # Origins for a gaussian distribution of rays.
            mean = [0,0,0]
            # Define a diagnonal covariance matrix.
            cov = [[self.param['xsize']**2, 0, 0], [0, self.param['ysize']**2, 0], [0, 0, self.param['zsize']**2]]
            sigma_to_fwhm = 2*np.sqrt(2*np.log(2))
            cov = np.array(cov)/sigma_to_fwhm**2
            x_offset, y_offset, z_offset = np.random.multivariate_normal(mean, cov, self.param['intensity']).T

        else:
            raise NotImplementedError(f"spatial_dist: {self.param['spatial_dist']} not implemented.")

        # Map x,y,z aligned offsets to the source origin and orientation.
        origin = (self.origin
                  + np.einsum('i,j', x_offset, self.xaxis)
                  + np.einsum('i,j', y_offset, self.yaxis)
                  + np.einsum('i,j', z_offset, self.zaxis))

        return origin

    def generate_direction(self, origin):
        normal = self.make_normal()
        D = self.random_direction(normal)
        return D

    def make_normal(self):
        array = np.empty((self.param['intensity'], 3))
        array[:] = self.param['zaxis']
        normal = array / np.linalg.norm(array, axis=1)[:, np.newaxis]
        return normal

    def random_direction(self, normal):

        spread = self.param['spread']
        dir_local = xicsrt_spread.vector_distribution(
            spread,
            self.param['intensity'],
            name=self.param['angular_dist'],
            )

        # Generate some basis vectors that are perpendicular to the normal.
        # For an isotropic distribution the direction does not matter.
        # for the xy distributions, the direction does matter.  This will
        # provide the correct behavior for a generic source with the normal
        # directed along the zaxis.
        o_1 = np.cross(normal, self.param['xaxis']) + np.cross(normal, self.param['zaxis'])
        o_1 /=  np.linalg.norm(o_1, axis=1)[:, np.newaxis]
        o_2  = np.cross(normal, o_1)
        o_2 /=  np.linalg.norm(o_2, axis=1)[:, np.newaxis]

        R = np.empty((self.param['intensity'], 3, 3))
        R[:,0,:] = o_2
        R[:,1,:] = o_1
        R[:,2,:] = normal
        
        direction = np.einsum('ij,ijk->ik', dir_local, R)
        return direction

    def generate_wavelength(self, direction):
        wtype = str.lower(self.param['wavelength_dist'])
        if wtype == 'monochrome':
            wavelength  = np.ones(self.param['intensity'], dtype = np.float64)
            wavelength *= self.param['wavelength']
        elif wtype == 'uniform':
            wavelength = np.random.uniform(
                self.param['wavelength_range'][0]
                ,self.param['wavelength_range'][1]
                ,self.param['intensity']
                )
        elif wtype == 'voigt':
            #random_wavelength = self.random_wavelength_normal
            #random_wavelength = self.random_wavelength_cauchy
            random_wavelength = self.random_wavelength_voigt
            wavelength = random_wavelength(self.param['intensity'])
        else:
            raise Exception(f'Wavelength distribution {wtype} unknown')

        if not np.all((self.param['velocity'] == 0.0)):
            # Doppler shift for each ray.
            c = const.physical_constants['speed of light in vacuum'][0]
            wavelength *= 1 - (np.einsum('j,ij->i', self.param['velocity'], direction) / c)

        return wavelength

    def random_wavelength_voigt(self, size=None):
        #Units: wavelength (angstroms), natural_linewith (1/s), temperature (eV)
        
        # Check for the trivial case.
        if (self.param['linewidth']  == 0.0 and self.param['temperature'] == 0.0):
            return np.ones(size)*self.param['wavelength']

        # Check for the Gaussian case.
        if (self.param['linewidth']  == 0.0):
            return self.random_wavelength_normal(size)

        # Check for zero temperature Voigt (Lorentzian case)
        if (self.param['temperature'] == 0.0):
            # I need to update the cauchy routine first.
            #raise NotImplementedError('Random Lorentzian distribution not implemented.')

            # TEMPORARY:
            # The voigt distribution generator cannot handle a zero temperature, so just add 1eV for now.
            self.param['temperature'] += 1.0

        c = const.physical_constants['speed of light in vacuum'][0]
        amu_kg = const.physical_constants['atomic mass unit-kilogram relationship'][0]
        ev_J = const.physical_constants['electron volt-joule relationship'][0]
        
        # Natural line width.
        gamma = (self.param['linewidth'] * self.param['wavelength']**2 / (4 * np.pi * c * 1e10))

        # Doppler broadened line width.
        sigma = (np.sqrt(self.param['temperature'] / self.param['mass_number'] / amu_kg / c**2 * ev_J)
                  * self.param['wavelength'] )

        rand_wave  = xicsrt_voigt.voigt_random(gamma, sigma, size)
        rand_wave += self.param['wavelength']
        return rand_wave

    def random_wavelength_normal(self, size=None):
        #Units: wavelength (angstroms), temperature (eV)
        c       = const.physical_constants['speed of light in vacuum'][0]
        amu_kg  = const.physical_constants['atomic mass unit-kilogram relationship'][0]
        ev_J    = const.physical_constants['electron volt-joule relationship'][0]
        
        # Doppler broadened line width.
        sigma = ( np.sqrt(self.param['temperature'] / self.param['mass_number'] / amu_kg / c**2 * ev_J)
                  * self.param['wavelength'] )

        rand_wave = np.random.normal(self.param['wavelength'], sigma, size)
        return rand_wave
    
    def random_wavelength_cauchy(self, size=None):
        # This function needs to be updated to use the same definitions
        # as random_wavelength_voigt.
        #
        # As currently writen natual_linewidth is not used in a way
        # consistent with physical units.
        #
        # It also may make sense to add some sort of cutoff here.
        # the extreme tails of the distribution are not really useful
        # for ray tracing.
        fwhm = self.param['linewidth']
        rand_wave = cauchy.rvs(loc=self.param['wavelength'], scale=fwhm, size=size)
        return rand_wave
    
    def generate_weight(self):
        # Weight is not currently used within XICSRT but might be useful
        # in the future.
        w = np.ones((self.param['intensity']), dtype=np.float64)
        return w
    
    def generate_mask(self):
        m = np.ones((self.param['intensity']), dtype=np.bool_)
        return m
    
    def ray_filter(self, rays):
        for filter in self.filter_objects:
            rays = filter.filter(rays)
        return rays



# -*- coding: utf-8 -*-
"""
Authors
-------
  - Novimir A. Pablant <nablant@pppl.gov>
  - Yevgeniy Yakusevich <eugenethree@gmail.com>
"""

import logging
import numpy as np

from xicsrt.util import profiler
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.sources._XicsrtPlasmaGeneric import XicsrtPlasmaGeneric
from xicsrt.tools.xicsrt_math import cyl_from_car

@dochelper
class  XicsrtPlasmaCylindrical(XicsrtPlasmaGeneric):
    """
    A cylindrical plasma oriented along the Y axis.

    .. Warning::
      This class is broken and out of date and needs to be updated.

    This class is meant only to be used as an example for generating
    more complicated classes for specific plasmas.

    plasma normal           = absolute X
    plasma x orientation    = absolute Z
    plasma y orientation    = absolute Y
    """
                
    def bundle_generate(self, bundle_input):
        #create a long list containing random points within the cube's dimensions
        x_offset = np.random.uniform(-1 * self.param['xsize']/2,  self.param['xsize']/2,  self.param['bundle_count'])
        y_offset = np.random.uniform(-1 * self.param['ysize']/2, self.param['ysize']/2, self.param['bundle_count'])
        z_offset = np.random.uniform(-1 * self.param['zsize']/2,  self.param['zsize']/2,  self.param['bundle_count'])
                 
        bundle_input['origin'][:] = (
            self.origin
            + np.einsum('i,j', x_offset, self.xaxis)
            + np.einsum('i,j', y_offset, self.yaxis)
            + np.einsum('i,j', z_offset, self.zaxis))       

        #convert from cartesian coordinates to cylindrical coordinates [radius, azimuth, height]
        radius, azimuth, height = cyl_from_car(np.array([z_offset, x_offset, y_offset]))
        
        # Let plasma temperature and emissivity fall off as a function of
        # radius.
        bundle_input['emissivity'][:]  = self.emissivity / radius
        bundle_input['temperature'][:] = self.temperature / radius
        bundle_input['velocity'][:]  = self.velocity
        
        return bundle_input
    



# -*- coding: utf-8 -*-
"""
Authors
-------
- Novimir A. Pablant <npblant@pppl.gov>
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.objects._ConfigObject import ConfigObject
from xicsrt.objects._GeometryObject import GeometryObject

@dochelper
class XicsrtBundleFilter(GeometryObject):
    """
    A base class for bundle filters.
    """
    
    def default_config(self):
        config = super().default_config()
        return config
    
    def filter(self, bundle_input):
        """
        This is the main filtering method that must be reimplemented
        for specific filter objects.
        """
        return bundle_input



# -*- coding: utf-8 -*-
"""
.. Authors:
    Conor Perks <cjperks@psfc.mit.edu>
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`XicsrtOpticCylindricalCrystal` class.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractCrystal import InteractCrystal
from xicsrt.optics._ShapeCylinder import ShapeCylinder

@dochelper
class XicsrtOpticCylindricalCrystal(InteractCrystal, ShapeCylinder):
    """
    A cylindrical crystal optic.
    """

    pass


# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>

A set of miscellaneous tools used by XICSRT.

Programming Notes
-----------------
Many of the functions have rather specific behavior for use within
xicsrt and may not have behavior that is well suited for general use outside
of their intended purposes.
"""

import numpy as np


def _convert_to_numpy(obj, inplace=False):
    """
    Convert any numerical lists in a dictionary to numpy arrays.

    This function is specifically meant to help with saving and loading
    of config dictionaries files and should not be considered a general tool.
    """
    if not inplace:
        obj = obj.copy()

    if isinstance(obj, dict):
        obj_keys = obj.keys()
    elif isinstance(obj, list):
        obj_keys = range(len(obj))
    else:
        raise TypeError('Object must be either a dict or a list.')

    for key in obj_keys:
        if isinstance(obj[key], list):
            # Don't convert empty lists.
            if obj[key]:
                new = np.array(obj[key])

                if new.dtype.char == 'U':
                    # Don't convert unicode lists to numpy.
                    # These cause issues with hdf5 export.
                    pass
                elif new.dtype.char == 'O':
                    obj[key] = _convert_to_numpy(obj[key])
                else:
                    obj[key] = new
        elif isinstance(obj[key], dict):
            obj[key] = _convert_to_numpy(obj[key])
    return obj


def _convert_from_numpy(obj, inplace=False):
    """
    Convert any numpy arrays in a dictionary to lists.

    This function is specifically meant to help with saving and loading
    of config dictionaries and should not be considered a general tool.
    """
    if not inplace:
        obj = obj.copy()

    if isinstance(obj, dict):
        obj_keys = obj.keys()
    elif isinstance(obj, list):
        obj_keys = range(len(obj))
    else:
        raise TypeError('Object must be either a dict or a list.')

    for key in obj_keys:
        if isinstance(obj[key], np.ndarray):
            obj[key] = obj[key].tolist()
        elif isinstance(obj[key], dict) or isinstance(obj[key], list):
            obj[key] = _convert_from_numpy(obj[key])
    return obj


def _debug_types(obj, level=0):
    """
    A debugging function to print data types within a nested data structure.
    """
    if isinstance(obj, dict):
        obj_keys = obj.keys()
    elif isinstance(obj, list):
        obj_keys = range(len(obj))
    else:
        raise TypeError('Object must be either a dict or a list.')

    for key in obj_keys:
        print(f"{' '*level} {key!s:15.15s} : {type(obj[key])}")
        if isinstance(obj[key], dict) or isinstance(obj[key], list):
            _debug_types(obj[key], level+1)


def _replace_strings(obj, old, new, inplace=False):
    """
    Recursively replaces all strings in the given object.

    This function is specifically meant to help with saving and loading
    of config dictionaries and should not be considered a general tool.
    """
    if not inplace:
        obj = obj.copy()

    if isinstance(obj, dict):
        obj_keys = obj.keys()
    elif isinstance(obj, list):
        obj_keys = range(len(obj))
    else:
        raise TypeError('Object must be either a dict or a list.')

    for key in obj_keys:
        if isinstance(obj[key], str):
            obj[key] = obj[key].replace(old, new)
        elif isinstance(obj[key], dict) or isinstance(obj[key], list):
            obj[key] = _replace_strings(obj[key], old, new)

    return obj